#!/usr/bin/env python

from palmiche.utils import tools
from string import ascii_uppercase
import numpy as np
"""
Taken from https://charmm-gui.org/?doc=lecture&module=scientific&lesson=7
"""

class Atom:
    """
    https://www.wwpdb.org/documentation/file-format-content/format33/sect9.html#ATOM    
    """
    def __init__(self, line):
        self.code = "ATOM"
        self.serial = int(line[6:11])
        self.name = line[12:16].strip()
        self.altLoc = line[16]
        self.resName = line[17:21].strip()
        self.chainID = line[21]
        self.resSeq = int(line[22:26])
        self.iCode = line[26]
        self.x = float(line[30:38])
        self.y = float(line[38:46])
        self.z = float(line[46:54])
        self.occupancy = line[54:60].strip()
        self.tempFactor = line[60:66].strip()
        self.element = line[76:78].strip()
        self.charge = line[78:80].strip()
        if self.occupancy: self.occupancy = float(self.occupancy)
        if self.tempFactor: self.tempFactor = float(self.tempFactor)

    def __getitem__(self, key):
        return self.__dict__[key]


class Hetatm:
    """
    https://www.wwpdb.org/documentation/file-format-content/format33/sect9.html#ATOM    
    """
    def __init__(self, line):
        self.code = "HETATM"
        self.serial = int(line[6:11])
        self.name = line[12:16].strip()
        self.altLoc = line[16]
        self.resName = line[17:21].strip()
        self.chainID = line[21]
        self.resSeq = int(line[22:26])
        self.iCode = line[26]
        self.x = float(line[30:38])
        self.y = float(line[38:46])
        self.z = float(line[46:54])
        self.occupancy = line[54:60].strip()
        self.tempFactor = line[60:66].strip()
        self.element = line[76:78].strip()
        self.charge = line[78:80].strip()
        if self.occupancy: self.occupancy = float(self.occupancy)
        if self.tempFactor: self.tempFactor = float(self.tempFactor)

    def __getitem__(self, key):
        return self.__dict__[key]


class CRYST1:
    """
    https://www.wwpdb.org/documentation/file-format-content/format33/sect8.html#CRYST1    
    """
    def __init__(self, line):
        self.a = float(line[6:15])			    #Real(9.3)     a              a (Angstroms).
        self.b = float(line[15:24])			    #Real(9.3)     b              b (Angstroms).
        self.c = float(line[24:33])			    #Real(9.3)     c              c (Angstroms).
        self.alpha = float(line[33:40])			#Real(7.2)     alpha          alpha (degrees).
        self.beta = float(line[40:47])			#Real(7.2)     beta           beta (degrees).
        self.gamma = float(line[47:54])			#Real(7.2)     gamma          gamma (degrees).
        self.sGroup = line[55:66]			    #LString       sGroup         Space  group.
        try:
            self.z = int(line[66:70])			    #Integer       z              Z value.
        except:
               self.z = "" 

    def __getitem__(self, key):
        return self.__dict__[key]
    def write(self):
        string = "CRYST1%9.3f%9.3f%9.3f%7.2f%7.2f%7.2f%-12s%4s\n"%\
            (self.a,self.b,self.c,self.alpha,self.beta,self.gamma,self.sGroup,self.z)
        return string
class Residue:
    """
    This will convert a list of Atom objects that belongs and will create a residue class taken into account the information of the first atom. It supoouse that all of them belong to the same residue of
    """
    def __init__(self, atoms):
        self.resName = atoms[0].resName
        self.resSeq = atoms[0].resSeq
        self.chainID = atoms[0].chainID
        self.atoms = atoms

    def __getitem__(self, key):
        return self.__dict__[key]

class PDB:
    def __init__(self, file):
        self.file = file
        self.title = ""
        self.cryst1 = CRYST1("0"*54+"  P 1 21 1     1")#Just to initialized to a generci cell, see the web referenced in CRYST1
        self.atoms = []
        self.parse()


    def parse(self):
        MODEL = None
        
        f = open(self.file, 'r')
        
        for line in f.readlines():

            if line.startswith('TITLE'): self.title = line.strip()
            if line.startswith('CRYST1'): self.cryst1 = CRYST1(line)
            if line.startswith('MODEL'): MODEL = int(line.split()[1])
            
            if line.startswith('ATOM'):
                atom = Atom(line)
                atom.MODEL = MODEL
                self.atoms.append(atom)
            if line.startswith('HETATM'):
                atom = Hetatm(line)
                atom.MODEL = MODEL
                self.atoms.append(atom)
        f.close()
        


    def get_chains(self):
        chains = set()
        for atom in self.atoms:
            chains.add(atom.chainID)
        return sorted(chains)#This atomathically convert to list
        

    def get_atoms(self):
        """Return a list of all atoms.
        each atom is represented as a dictionary.
        """
        return [x.__dict__ for x in self.atoms]
        

    def get_model(self, model_num, to_dict=True):
        """Return all atoms where MODEL == model_num"""
        model_atoms = [x for x in self.atoms if x.MODEL == model_num]
        if to_dict:
            return [atom.__dict__ for atom in model_atoms]
        else:
            return model_atoms
    

    def get_residues(self):
        """Rearrange the data as residues function.

        Returns:
            list: a list of Residues objects
        """
        residues = []
        i = 0
        while i < len(self.atoms):
            resAtoms = []
            resName = self.atoms[i].resName
            resSeq = self.atoms[i].resSeq
            chainID = self.atoms[i].chainID
            j = i
            while self.atoms[j].resName == resName and self.atoms[j].resSeq == resSeq and self.atoms[j].chainID == chainID:
                resAtoms.append(self.atoms[j])
                j += 1
                if j >= len(self.atoms): break
            residues.append(Residue(resAtoms))
            i = j
        
        return residues


    def edges(self):
        x_min = min(self.atoms, key=lambda item: item.x)
        y_min = min(self.atoms, key=lambda item: item.y)
        z_min = min(self.atoms, key=lambda item: item.z)

        x_max = max(self.atoms, key=lambda item: item.x)
        y_max = max(self.atoms, key=lambda item: item.y)
        z_max = max(self.atoms, key=lambda item: item.z)
        
        length = [round(x_max.x-x_min.x, 3), round(y_max.y-y_min.y, 3), round(z_max.z-z_min.z, 3)]
        
        print(f"""
X axes:
       Length: {length[0]} Angstroms
       Bounding atoms: {x_max.name}_{x_max.serial} ({x_max.resName}_{x_max.resSeq}-{x_max.chainID}) ----- {x_min.name}_{x_min.serial} ({x_min.resName}_{x_min.resSeq}-{x_min.chainID})
Y axes:
       Length: {length[1]} Angstroms
       Bounding atoms: {y_max.name}_{y_max.serial} ({y_max.resName}_{y_max.resSeq}-{y_max.chainID}) ----- {y_min.name}_{y_min.serial} ({y_min.resName}_{y_min.resSeq}-{y_min.chainID})
Z axes:
       Length: {length[2]} Angstroms
       Bounding atoms: {z_max.name}_{z_max.serial} ({z_max.resName}_{z_max.resSeq}-{y_max.chainID}) ----- {z_min.name}_{z_min.serial} ({z_min.resName}_{z_min.resSeq}-{z_min.chainID})
            """)
        return tuple(length)
        
    def box_center(self, resName_selection = [], H_atoms = True):
        if resName_selection:
            atoms = [atom for atom in self.atoms if atom.resName in resName_selection]
        else:
            atoms = self.atoms
        
        if not H_atoms:
            atoms = [atom for atom in atoms if atom.name[0] not in 'Hh']
        
        x_min = min(atoms, key=lambda item: item.x)
        y_min = min(atoms, key=lambda item: item.y)
        z_min = min(atoms, key=lambda item: item.z)

        x_max = max(atoms, key=lambda item: item.x)
        y_max = max(atoms, key=lambda item: item.y)
        z_max = max(atoms, key=lambda item: item.z)
         
        return round((x_max.x+x_min.x)/2,3), round((y_max.y+y_min.y)/2,3), round((z_max.z+z_min.z)/2,3)
    
    def centroid(self, resName_selection = [], H_atoms = True):
        if resName_selection:
            atoms = [atom for atom in self.atoms if atom.resName in resName_selection]
        else:
            atoms = self.atoms
        
        if not H_atoms:
            atoms = [atom for atom in atoms if atom.name[0] not in 'Hh']
        
        centroid = (sum([np.array([atom.x,atom.y,atom.z]) for atom in atoms]) / len(atoms))
        centroid = [round(coord, 3) for coord in centroid]
        return tuple(centroid)


    def fix_chains(self, HETATM = False):
        #In the case that you ahve only one aa per chain with the same residue
        #number this will fail. It will asign chain A to all. 
        
        #Here we are using only the ATOMS for the CHAIN rectificatin, not the HETATM
        #If HETATM False, use only ATOMS,
        #If HETATM True, use ATOMS and HEATM
        if not HETATM:
            atoms = [atom for atom in self.atoms if atom.code == "ATOM"]
        else:
            atoms = self.atoms
        
        chain = 0    
        for (i, atom) in enumerate(atoms):
            if i == 0:
                atom.chainID = ascii_uppercase[chain]
            elif atoms[i].resSeq >= atoms[i-1].resSeq:
                atom.chainID = ascii_uppercase[chain]
            else:
                chain += 1
            atom.chainID = ascii_uppercase[chain]
    
    def add_element(self):
        """
        This function add the columnd elemnt to the pdb data

        Returns
        -------
        None.

        """
        for atom in self.atoms:
            if not atom.element:
                element = ""
                for char in atom.name:
                    if char.isalpha():
                        element = char.upper()
                        break
                atom.element = element
    

    def write(self, out_file = None, backup = True):
        #This could be used if some modifications were made in the instance
        if out_file:
            to_work = out_file
        else:
            to_work = self.file
            if backup: tools.backoff(to_work)
        with open(to_work,'w') as out:
            if self.title: out.write(self.title+'\n')
            if self.cryst1.a or self.cryst1.b or self.cryst1.c: #If the length of at least one vector is greater than 0
                out.write(self.cryst1.write())
            if self.atoms[0].MODEL != None: out.write("MODEL %6i\n"%(self.atoms[0].MODEL))
            
            
            """
            For the case that the protein have more than one chain, teh TER 
            flag will be printed. If not (thae protein has only one chain or 
            the chain is not decleared) the TER flag will not printed
            """
            CHAINS_ID = self.get_chains()#Becasue get_chains belongs to the class you need to call it like a self.function(), the method of the class
            if len(CHAINS_ID) > 1:
                last = self.atoms[0].chainID
                for (i,atom) in enumerate(self.atoms):
                    
                    #INCREDIBLE, PropKa if the atoms name is shifted to the left dont understand!!!
                    #For that reason I need to change when the len is 3. NO IDEA!!!!
                    if len(atom.name) == 3:
                        string_name = f"{atom.name:>4}"
                    else:
                        string_name = f"{atom.name:^4}"
                        
                    if atom.chainID != last:
                        out.write("TER\n")
                        last = atom.chainID

                    #This check is necesary for the case that some residues has a name greater than 3 letters (ASPH)
                    #This generate something like ASPHA, where A is the Chain identifier
                    
                    out.write(f"{atom.code:<6}{atom.serial:5} {string_name}"
                              f"{atom.altLoc:1}{atom.resName:<4}{atom.chainID:1}"
                              f"{atom.resSeq:4}{atom.iCode:1}   {atom.x:8.3f}"
                              f"{atom.y:8.3f}{atom.z:8.3f}{atom.occupancy:6.2f}"
                              f"{atom.tempFactor:6.2f}           {atom.element:2}"
                              f"{atom.charge:2}\n")
                out.write("TER\n")#This is for the last TER flag
            else:
                for (i,atom) in enumerate(self.atoms):
                    #INCREDIBLE, PropKa if the atoms name is shifted to the left dont understand!!!
                    if len(atom.name) == 3:
                        string_name = f"{atom.name:>4}"
                    else:
                        string_name = f"{atom.name:^4}"
                        
                    out.write(f"{atom.code:<6}{atom.serial:5} {string_name}"
                              f"{atom.altLoc:1}{atom.resName:<4}{atom.chainID:1}"
                              f"{atom.resSeq:4}{atom.iCode:1}   {atom.x:8.3f}"
                              f"{atom.y:8.3f}{atom.z:8.3f}{atom.occupancy:6.2f}"
                              f"{atom.tempFactor:6.2f}           {atom.element:2}"
                              f"{atom.charge:2}\n")
                        

if __name__ == '__main__':
    #import sys
    #pdb = PDB(sys.argv[1])
    #pdb = PDB('PROTEIN.pdb')
    #print(pdb)
    #atoms = pdb.atoms
    #print(atoms[0].__dict__)
    #print(pdb.get_chains())
    #for atom in atoms: #Could be used to rectificate the pdb and put the chain information
    #    atom.chainID = "G" 
    #print(pdb.get_chains())
    #pdb.cryst1.a = 500
    #print(pdb.cryst1.write())
    #for atom in atoms:
    #    if atom.resName == "LIA":
    #        atom.x = 0
    #        atom.y = 0
    #        atom.z = 0
    #pdb.write("j.pdb")
    #print(isinstance(pdb, PDB))
    #print(type(pdb).__name__)
    

    #print the name of each atom
    #for atom in atoms:
    #    print(atom.serial, atom.name, atom.x, atom.y, atom.z, atom.element)
    #lig_name = "HV6"
    #for atom in pdb.atoms:
    #    if atom.resName == lig_name:
    #        atom.resName = f"LI{atom.chainID}"
    #print(pdb.centroid(), pdb.box_center(), pdb.edges())
    #rint(pdb.atoms[-1].__dict__)
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from palmiche.utils import tools, xvg
import tempfile
import os
import tqdm
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from scipy import integrate
import warnings

def AleWeights(temperatures, internal_energies, target_temperatures, integrator = 'trapezoid', alpha = None, weight_GROMACS_format = True):
    """
    Here explain the definition of the gromacs weight
    https://manual.gromacs.org/documentation/2019/reference-manual/algorithms/expanded-ensemble.html
    Tengo que mejorar lo del input del data frame es muy complicado dar ese input, que sea una arraytemperatura, el otro el valor de la energia, y em temp_list
    para las temperaturasw que quiero calcular los weights, de esta forma es mucho mas simple, tengo que ver el signo de la matriz porque sino tengo que cambiar el signo par lo del alpha
    MEJORAR!!!!
    Calculation of the weights for simulated tempering. The weights respond to the hamiltonian:
        H(p,q,T) = V(q) + K(p) - weight/beta; where beta = 1/KbT
    In GROMACS the hamiltonian is deffined:
        H(p,q,T) = V(q) + K(p) - weight_GROMACS
    Therefore:
        weight_GROMACS = weight/beta
    To convert to GROMACS weights you just need to provided weight_GROMACS_format = True (default option)

    Args:
        DataFrame (pd.DataFrame): A Data Frame with temperatures as index and one column with the internal energy with name total_energy.
        temp_list (list): The list of temperature to calculate the weights. At least two. If some temperature is not in the temperature range provided with DataFrame a ValueError will prompt.
        integrator (str, optional): Integration method. Are currently supported:
            trapezoid --> scipy.integrate.trapezoid
            trapz --> numpy.trapz (Default)
            simps --> scipy.integrate.simpson 
        weight_GROMACS_format (bool, optional): If True each weight will be divided by its corresponded beta value of temperature. Defaults to True.

    Returns:
        weights (numpy array): weights for each temperature in the order of temp_list. weights[0] will be the weight of temp_list[0] and so on.
    """

    # Selecting the integrator method
    if integrator == 'trapezoid':
        integrator = integrate.trapz
    elif integrator == 'trapz':
        integrator = np.trapz
    elif integrator == 'simpson':
        integrator = integrate.simps
    else:
        raise ValueError(f"The integration method '{integrator}' is not available.\n"\
            "Current possible selections are:\n"\
            "   -> trapezoid\n"\
            "   -> trapz\n"\
            "   -> simpson (Default)")
    target_temperatures = np.array(target_temperatures)
    # Converting to pandas.DataFrame and sorting the values respect to temperature
    DataFrame = pd.DataFrame(
        {
           'temperatures': temperatures,
           'internal_energies': internal_energies,
        }
    )
    DataFrame.sort_values(by='temperatures', inplace=True)
    DataFrame.reset_index(drop=True, inplace=True)
    # Converting target_temperatures to numpy array
    # Creating the integrand U(T)/T^2
    DataFrame['integrand'] = DataFrame.internal_energies / (DataFrame.temperatures**2)
    T_MAX = DataFrame.temperatures.max()
    T_MIN = DataFrame.temperatures.min()


    # Deffining the coefficient matrix
    M = 1*np.eye(len(target_temperatures), k = 1, dtype=int)
    np.fill_diagonal(M, -1)
    M[-1] = np.ones(len(target_temperatures), dtype = int)

    # Calculating solution vector, based on integral from Ta to Tb of 1/KB * U(T)/T^2
    # Calculating the first component depending if alpha was selected or not
    print('Calculating integrals for solution vector...')
    T0, T1 = target_temperatures[0], target_temperatures[1]
    if not T_MIN <= T0 <= T_MAX:
        warnings.warn(f"Temperature {T0} provided with temp_list is not in the interval of DataFrame; which is [{T_MIN}, {T_MAX}]")
    if not T_MIN <= T0 <= T_MAX:
        warnings.warn(f"Temperature {T1} provided with temp_list is not in the interval of DataFrame; which is [{T_MIN}, {T_MAX}]")
    # Selecting the interval of integration
    data = DataFrame[(DataFrame.temperatures >= T0) & (DataFrame.temperatures <= T1)]
    if len(data) < 4: # At least 4 points
        raise RuntimeError(f"There are not enough data points for the range of temperatures {T0} to {T1}")

    if alpha:
        gamma = alpha/(1-alpha)
        b = [-integrator(data.integrand, data.temperatures) / tools.CTE.Kb - np.log(gamma*(len(temperatures) - 1))]
    else:
        b = [-integrator(data.integrand, data.temperatures) / tools.CTE.Kb]

    for i in tqdm.tqdm(range(1, len(target_temperatures)), total =  len(target_temperatures) - 2):
        if i < len(target_temperatures) - 1:
            # Getting integration interval extrem values and checking that belongs to the DataFrame provided.
            T0, T1 = target_temperatures[i], target_temperatures[i+1]
            if not T_MIN <= T0 <= T_MAX:
                warnings.warn(f"Temperature {T0} provided with temp_list is not in the interval of DataFrame; which is [{T_MIN}, {T_MAX}]")
            if not T_MIN <= T0 <= T_MAX:
                warnings.warn(f"Temperature {T1} provided with temp_list is not in the interval of DataFrame; which is [{T_MIN}, {T_MAX}]")

            # Selecting the interval of integration
            data = DataFrame[(DataFrame.temperatures >= T0) & (DataFrame.temperatures <= T1)]
            if len(data) < 4: # At least 4 points
                raise RuntimeError(f"There are not enough data points for the range of temperatures {T0} to {T1}")

            b.append(-integrator(data.integrand, data.temperatures) / tools.CTE.Kb)
        else:
            # We reach the last temperature
            b.append(0)
            break
    # Calculate the weights:
    weights = np.linalg.solve(M, b)
     
    # Convert if needed to GROMACS weights
    if weight_GROMACS_format:
        minimum = weights.min()
        if minimum > 0:
            weights -= minimum
        else:
            weights += np.abs(minimum)
    
    return weights

def PandeWeights(temperatures, energies, alpha = None, weight_GROMACS_format = True):
    # Here explain the definition of the gromacs weight
    # https://manual.gromacs.org/documentation/2019/reference-manual/algorithms/expanded-ensemble.html
    # The use of the alpha is not completly right from a theretical point of view for this method
    assert len(temperatures) == len(energies), "temperatures and energies must have the same number of elements"
    # Deffining the coefficient matrix
    M = 1*np.eye(len(temperatures), k = 1, dtype=int)
    np.fill_diagonal(M, -1)
    M[-1] = np.ones(len(temperatures), dtype = int)
    # Solution Vector
    b = np.zeros(len(temperatures))
    # Initialize b[0] dependign if alpha was selected
    if alpha:
        gamma = alpha/(1-alpha)
        b[0] = 0.5*(tools.beta(temperatures[1]) - tools.beta(temperatures[0]))*(energies[0] + energies[1]) - np.log(gamma*(len(temperatures) - 1))
    else:
        b[0] = 0.5*(tools.beta(temperatures[1]) - tools.beta(temperatures[0]))*(energies[0] + energies[1])

    for i in range(1, len(energies)):
        try:
            b[i] = 0.5*(tools.beta(temperatures[i + 1]) - tools.beta(temperatures[i]))*(energies[i] + energies[i + 1])
        # I reached the last element of the list so I need to add cero to the array but I already initialized on zero.
        except:
            break
    # Solve the system of equations
    # Calculate the weights:
    weights = np.linalg.solve(M, b)
     
    # Convert if needed to GROMACS weights
    if weight_GROMACS_format:
        minimum = weights.min()
        if minimum > 0:
            weights -= minimum
        else:
            weights += np.abs(minimum)
    return weights

def get_energy(edr, annealing_times, energy_type = 'Potential', out_fig = 'energy_distribution.svg'): # Could be Total-Energy
    """The two input is a list that could be obtained from mdp.annealing3
    The structure is the following
    annealing_times = [t1, t2, t3, t4, t5, t6]
    These mean:
    From mdp.annealing, we also obtain the list
    annealing_temperatures = [T1, T1, T2, T2, T3, T3]
    Tis two list are connected and mean:
    from time t1 to t2 the temperature was constant, therefore from here we should get the average energy for T1.
    Then from t2 to t3 the temperature increase from T1 to T2
    Then from t3 to t4 the temperature is constant at T2,therefore from here we should get the average energy for T2.
    And so on and so for.

    Args:
        annealing_temperatures (_type_): _description_
        annealing_times (_type_): _description_
    """
    fig, ax = plt.subplots(figsize = (16,9))
    data = pd.DataFrame()
    xvg_tmp_file = tempfile.NamedTemporaryFile(suffix='.xvg')
    energy = []
    iterator = range(0, len(annealing_times)-1, 2)

    for state, index in tqdm.tqdm(enumerate(iterator), total=len(iterator)):#enumerate(iterator):# # the calculation is per pair of times, beetween the first to time the temperature was keep constant, then the system was heated and repeated again.
        run = tools.run(f"export GMX_MAXBACKUP=-1; echo {energy_type} | gmx energy -f {edr} -b {annealing_times[index]} -e {annealing_times[index + 1]} -o {xvg_tmp_file.name} | grep \'{energy_type.replace('-',' ')}\'")
        energy.append(float(run.stdout.split()[-5]))
        """
        Energy                      Average   Err.Est.       RMSD  Tot-Drift
        -------------------------------------------------------------------------------
        Potential                -1.30028e+06         --     1682.1   -2422.24  (kJ/mol)
        Total Energy                -952595         --    2606.81    -3688.3  (kJ/mol)
        """
        # Getting the histograms and checking for the same len in all intervals
        if state == 0:
            data[state] = xvg.XVG(xvg_tmp_file.name).data[:,1]
        else:
            xvg_data = xvg.XVG(xvg_tmp_file.name).data[:,1]
            if xvg_data.shape[0] > data.shape[0]:
                data[state] = xvg_data[:data.shape[0]]
            else:
                data = data.iloc[:xvg_data.shape[0]]
                data[state] = xvg_data


    print(data)
    sns.histplot(data = data, element='poly', stat = 'probability', axes = ax)
    ax.set(
        xlabel = f'{energy_type} [kJ/mol]',
        ylabel = 'Probability',
        title = f'Distribution of {energy_type}')
    # plt.show()
    fig.savefig(out_fig)
    return energy

def get_weights_from_log(log, plot = False):
    """This function read a log file obtained from simulated tempering and returns a numpy array
    of shape (number of time points, number of states, 2)
    The first axis of the array refears to each output on the log file.
    The second axis refears to the temperature states
    And the last one to the counts and value of the corresponded weights
    E.g., if we would like to know the value of the weight on the second output of the log file for the temperature state 5.
    >>> from palmiche.simutemp import weights
    >>> time, info_weights = weights.get_weights_from_log('tempering.log')
    >>> info_weights[1,4,1]
    387.69299

    Get the final weights:
    >>> info_weights[-1,:,1]
    array([   0.     ,   95.78638,  194.94769,  291.73962,  387.69299,
        484.79767,  563.36194,  609.72424,  656.0224 ,  702.54724,
        747.32648,  793.23291,  838.1333 ,  880.62061,  922.65308,
        966.77673, 1008.16711, 1051.17847, 1091.65161, 1133.02808,
       1175.10168, 1212.64014, 1249.82324, 1289.07568, 1329.84229,
       1366.30273])

    Args:
        log (str): path to the log file

    Returns:
        numpy.array: The weights information.
    """
    with open(log, 'r') as f:
        log_file = f.readlines()

    i = 0
    time = []
    weights_info = []
    weights_0 = []
    while i < len(log_file):
        if 'init-lambda-weights[' in log_file[i]:
            weights_0.append(float(log_file[i].split('=')[-1]))

        if 'MC-lambda information' in log_file[i]:
            # Finding the time
            for j in range(i,0,-1):
                if log_file[j].startswith('           Step           Time'):
                    j += 1
                    time.append(float(log_file[j].split()[-1]))
                    break
            # Finding the weight
            weights_info_tmp = []
            i += 3
            while log_file[i] != '\n':
                split = log_file[i].split()
                count = int(split[2])
                weight = float(split[3])
                weights_info_tmp.append((count, weight))
                i += 1
            weights_info.append(weights_info_tmp)
        i += 1
    # Add weights at t = 0, because the counts are all 0 and I delate the entrances with total count 0 in next lines,
    # What i could do is put 1 in the initial temperature
    time.insert(0,0)
    weights_info.insert(0,list(zip([1] + (len(weights_0) - 1)*[0], weights_0)))

    #Converting to array
    time = np.array(time)
    weights_info = np.array(weights_info)
    # Some times (I don't know why) GROMACS reset all the weights and all the counts are 0. We need to eliminate those points
    sum_of_weights = weights_info[:,:,0].sum(axis = 1)
    time = time[sum_of_weights != 0]
    weights_info = weights_info[sum_of_weights != 0]
    sum_of_weights = sum_of_weights[sum_of_weights != 0]


    if plot:
        dir = os.path.dirname(log)
        fig, axes = plt.subplots(2, figsize = (16,9), sharex=True)
        NUM_COLORS = weights_info.shape[1]
        cm = plt.get_cmap('viridis')#gist_rainbow viridis
        for axe in axes:
            axe.set_prop_cycle('color', [cm(1.*j/NUM_COLORS) for j in range(NUM_COLORS)])

        probability = weights_info[:,:,0] / sum_of_weights[:,np.newaxis]
        for j in range(weights_info.shape[1]):
            #axes[0].plot(time, weights_info[:,j,0], label = str(j))
            axes[0].plot(time, probability[:,j], label = str(j))
            axes[1].plot(time, weights_info[:,j,1])
        
        fig.legend(loc = 'lower center', ncol = int(weights_info.shape[1] / 2))
        axes[0].set(
            xlim = (time.min(), time.max()),
            ylim = (0,1),
            ylabel = 'Probability',
            )
        axes[1].set(
            xlabel = 'Time [ps]',
            ylabel = 'Weight values'
        )
        #plt.show()
        fig.savefig(os.path.join(dir,'weights_progression.svg'), bbox_inches="tight")

        # Plotting the violin plot of the weights
        df = pd.DataFrame()
        for j in range(weights_info.shape[1]):
            #df[temperatures[j]] = weights_info[:,j,1]
            df[j] = weights_info[:,j,1]
        # Set up the matplotlib figure
        sns.set_theme(style="whitegrid")
        fig, ax = plt.subplots(figsize=(25, 25))

        # Draw a violinplot with a narrower bandwidth than the default
        sns.violinplot(data=df, palette="Set3", bw=.2, cut=1, linewidth=1)
        # The plot is not over the actual temperatures, the temperatures ara only labels
        ax.plot(range(len(weights_info[0,:,1])), weights_info[0,:,1], '-o',  label = 'Initial weights')
        ax.set(
            title = 'Weights per state over the entire simulation',
            xlabel = 'Sate',
            ylabel = 'Weight',
        )
        plt.legend()
        sns.despine(left=True, bottom=True)
        #plt.show()
        fig.savefig(os.path.join(dir,'weights_per_state.svg'), bbox_inches="tight")
        sns.reset_defaults()

    return time, weights_info


if __name__ == '__main__':
    print("You are in the submodule weights of palmiche.simutemp")
    # a = get_weights_from_log('/home/ale/mnt/smaug/TEST/500_H2O/tempering.log')
    # print(a)

    # o = get_energy('/home/ale/mnt/smaug/TEST/500_H2O/tempering.edr', [0,20,25,30,50,200, 230,240,563,600,700,800,900,950])
    # print(o)

    # a = PandeWeights([100,400,500], [20,17,256], alpha = 0.0001)
    # print(a)
    # temperatures = list(np.linspace(300, 501, 400))
    # energies = range(len(temperatures))
    # target_temperatures = [300, 325, 350, 375, 400, 425, 450, 475, 500]
    # a = AleWeights(temperatures, energies, target_temperatures, alpha = 2)
    # print(a)
    # path = '/home/ale/mnt/smaug/TEST/water/AleWeights'#'/home/ale/mnt/smaug/TEST/500_H2O/'#'/home/ale/mnt/smaug/TEST/water/AleWeights'
    # temp_list = range(300, 332, 2)
    # temperature = xvg.XVG(os.path.join(path,'temperature.xvg')).data
    # total_energy = xvg.XVG(os.path.join(path,'total_energy.xvg')).data
    # # mask = (temperature[:,1] >= min(temp_list) - 2) & (temperature[:,1] <= max(temp_list) + 2)
    # # temperature = np.array(temperature[mask])
    # # total_energy = np.array(total_energy[mask])
    # print(temperature[:,1])
    # A = AleWeights(
    #     temperature[:,1],
    #     total_energy[:,1],
    #     temp_list,
    #     alpha = None,
    #     weight_GROMACS_format = True,
    #     integrator = 'trapezoid')
    # print(A)
    _, w = get_weights_from_log('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_P/7e27/sys_MMV007839_Cell_891_SP_param/windows/00085/production.log', plot = True) # AleWeights


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import socket, glob
sbatch_translation = {
    #https://slurm.schedmd.com/sbatch.html
    'A':'account',
    'account':'account',
    'acctg-freq':'acctg-freq',
    'a':'array','array':'array',
    'batch':'batch',
    'bb':'bb',
    'bbf':'bbf',
    'b':'begin', 'begin': 'begin',
    'D':'chdir', 'chdir':'chdir',
    'cluster-constraint':'cluster-constraint',
    'M':'clusters', 'clusters':'clusters',
    'comment':'comment',
    'C':'constraint', 'constraint':'constraint',
    'container':'container',
    'contiguous':'contiguous',
    'S':'core-spec', 'core-spec':'core-spec',
    'cores-per-socket':'cores-per-socket',
    'cpu-freq':'cpu-freq',
    'cpus-per-gpu':'cpus-per-gpu',
    'c':'cpus-per-task', 'cpus-per-task':'cpus-per-task',
    'deadline':'deadline',
    'delay-boot':'delay-boot',
    'd':'dependency', 'dependency':'dependency',
    'm':'distribution', 'distribution':'distribution',
    'e':'error', 'error':'error',
    'x':'exclude', 'exclude':'exclude',
    'exclusive':'exclusive',
    'export':'export',
    'export-file':'export-file',
    'B':'extra-node-info', 'extra-node-info':'extra-node-info',
    'get-user-env':'get-user-env',
    'gid':'gid',
    'gpu-bind':'gpu-bind',
    'gpu-freq':'gpu-freq',
    'G':'gpus', 'gpus':'gpus',
    'gpus-per-node':'gpus-per-node',
    'gpus-per-socket':'gpus-per-socket',
    'gpus-per-task':'gpus-per-task',
    'gres':'gres',
    'gres-flags':'gres-flags',
    'h':'help', 'help':'help',
    'hint':'hint',
    'H':'hold', 'hold':'hold',
    'ignore-pbs':'ignore-pbs',
    'i':'input', 'input':'input',
    'J':'job-name', 'job-name':'job-name',
    'kill-on-invalid-dep':'kill-on-invalid-dep',
    'L':'licenses', 'licenses':'licenses',
    'mail-type':'mail-type',
    'mail-user':'mail-user',
    'mcs-label':'mcs-label',
    'mem':'mem',
    'mem-bind':'mem-bind',
    'mem-per-cpu':'mem-per-cpu',
    'mem-per-gpu':'mem-per-gpu',
    'mincpus':'mincpus',
    'network':'network',
    'nice':'nice',
    'k':'no-kill', 'no-kill':'no-kill',
    'no-requeue':'no-requeue',
    'F':'nodefile', 'nodefile':'nodefile',
    'w':'nodelist','nodelist':'nodelist',
    'N':'nodes', 'nodes':'nodes',
    'n':'ntasks', 'ntasks':'ntasks',
    'ntasks-per-core':'ntasks-per-core',
    'ntasks-per-gpu':'ntasks-per-gpu',
    'ntasks-per-node':'ntasks-per-node',  
    'ntasks-per-socket':'ntasks-per-socket',    
    'open-mode':'open-mode',
    'o':'output', 'output':'output',
    'O':'overcommit', 'overcommit':'overcommit',
    's':'oversubscribe', 'oversubscribe':'oversubscribe',
    'parsable':'parsable',
    'p':'partition', 'partition':'partition',
    'power':'power',    
    'priority':'priority',    
    'profile':'profile',
    'propagate':'propagate',
    'q':'qos', 'qos':'qos',
    'Q':'quiet','quiet':'quiet',
    'reboot':'reboot',
    'requeue':'requeue',
    'reservation':'reservation',
    'signal':'signal',  
    'sockets-per-node':'sockets-per-node',
    'spread-job':'spread-job',
    'switches':'switches',
    'test-only':'test-only',
    'thread-spec':'thread-spec',
    'threads-per-core':'threads-per-core',
    't':'time', 'time':'time', #Acceptable time formats include "minutes", "minutes:seconds", "hours:minutes:seconds", "days-hours", "days-hours:minutes" and "days-hours:minutes:seconds".
    'tmp':'tmp',
    'uid':'uid',
    'usage':'usage',
    'use-min-nodes':'use-min-nodes',
    'v':'verbose', 'verbose':'verbose',
    'V':'version', 'version':'version',
    'W':'wait', 'wait':'wait',
    'wait-all-nodes':'wait-all-nodes',
    'wckey':'wckey',
    'wrap':'wrap',
}
class JOB:
    def __init__(self, sbatch_keywords = None, GROMACS_version = 2021.4, mdrun_keywords = None, build_GROMACS_section = None) -> None:
        self.hostname = socket.gethostname()
        self.GROMACS_version = GROMACS_version
        self.build_GROMACS_section = build_GROMACS_section
        self.default_times = {
            'deflt': '2-00:00:00',
            'long': '5-00:00:00',
            'debug': '30:00',
            'medium': '2-00:00:00',
            'fat+': '2-00:00:00',
            'fat': '2-00:00:00',
            'int': '2-00:00:00',
            'gpu': '2-00:00:00',
            'gpu:test': '30:00',
            'gpu-hub':'2-00:00:00',
        }
        self.sbatch_keywords = {
            'job-name': 'run',
            'output': 'myjob.out',
            'error': 'myjob.err',
            'nice': 0,
            'nodes': 1,
            'gpus':1,
        }
        # This part need to be improved
        if self.hostname == 'smaug' or self.hostname.startswith('fang'):
            self.sbatch_keywords['cpus-per-task'] = 12
            self.sbatch_keywords['partition'] = 'deflt'
        elif self.hostname.startswith('gwd'): # This is quiet lazy, but is because I don't know the names of the gwdg nodes
            self.sbatch_keywords['cpus-per-task'] = 10
            self.sbatch_keywords['partition'] = 'gpu-hub'#'medium'#'gpu-hub'
            self.sbatch_keywords['exclude'] = 'gwdo[161-180]'
            # self.sbatch_keywords['account'] = 'all'
            # self.sbatch_keywords['constraint'] = 'scratch'
        else:
            self.sbatch_keywords['cpus-per-task'] = 12
            self.sbatch_keywords['partition'] = 'deflt'
         
        # Translating the user sbatch_keywords in case that something was provided
        if sbatch_keywords:       
            sbatch_keywords_translated = {}
            for key in sbatch_keywords:
                try:
                    sbatch_keywords_translated[sbatch_translation[key]] = sbatch_keywords[key]
                except:
                    raise Exception(f'The keyword {key} of sbatch_keywords is not a valid sbatch keyword argument')       
            
            # Update with the user sbatch keword provided
            self.sbatch_keywords.update(sbatch_keywords_translated)
            
            # Setting the correct time
            if 'time' not in sbatch_keywords_translated:
                self.sbatch_keywords['time'] = self.default_times[self.sbatch_keywords['partition']]

        # Deffining default options. In case that flags was provided, the value must be a bool type
        self.mdrun_keywords = {
            'nt': self.sbatch_keywords['cpus-per-task'], # This is not so straightforwart. Because ntasks and cpus-per-task, but is a default option, you could just provided a new value
            'cpi': True,
            'stepout': 5000,
            'v':True,
        }
        # Updating the values of self.mdrun_keywords with the provided by the user
        if mdrun_keywords:        
            self.mdrun_keywords.update(mdrun_keywords)
        
    # Now we will start to construct the file
    def string(self):
        job_str = '#!/bin/bash\n'
        if self.hostname == 'smaug' or self.hostname.startswith('fang') or self.hostname.startswith('gwd'):
            # SBATCH section
            for key in self.sbatch_keywords:
                job_str += f"#SBATCH --{key}={self.sbatch_keywords[key]}\n"
            # Echo some important variables
            job_str += (
                "\n\n# This block is echoing some SLURM variables\n"\
                "echo \"Job execution start: $(date)\"\n"\
                "echo \"JobID = $SLURM_JOBID\"\n"\
                "echo \"Host = $SLURM_JOB_NODELIST\"\n"\
                "echo \"Jobname = $SLURM_JOB_NAME\"\n"\
                "echo \"Subcwd = $SLURM_SUBMIT_DIR\"\n"\
                "echo \"SLURM_TASKS_PER_NODE = $SLURM_TASKS_PER_NODE\"\n"\
                "echo \"SLURM_CPUS_PER_TASK = $SLURM_CPUS_PER_TASK\"\n"\
                "echo \"SLURM_CPUS_ON_NODE = $SLURM_CPUS_ON_NODE\"\n\n"\
                )
            # Loading modules
            if self.hostname == 'smaug' or self.hostname.startswith('fang'):
                job_str += (
                        f"source /data/shared/opt/gromacs/{self.GROMACS_version}/bin/GMXRC\n"\
                        f"PATH=$PATH:\"/data/shared/opt/gromacs/{self.GROMACS_version}/bin\"\n"\
                        f"#export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:\n\n"
                        )
            # This could give some problems for testing GROMACS versions
            elif self.hostname.startswith('gwd'):
                job_str += 'source /usr/users/cmb/shared/spack-0.17/shared.bash\n'\
                'if [[ $HOSTNAME == gwdo* ]]; then\n'\
                f'  module load  ivybridge/gromacs@{self.GROMACS_version}\n'\
                'else\n'\
                f'  module load gromacs@{self.GROMACS_version}\n'\
                'fi\n\n'\
        
        job_str += 'cd $(pwd)\n\n'
        # Constructing GROMACS command
        if self.build_GROMACS_section:
            job_str += self.build_GROMACS_section
        else:
            if 's' not in self.mdrun_keywords and 'deffnm' not in self.mdrun_keywords:
                tpr_name = glob.glob('*.tpr')[0].split('.')[0]
                CMD_GROMACS = f'gmx mdrun -v -deffnm {tpr_name}'
            else:
                CMD_GROMACS = 'gmx mdrun'
                try:
                    tpr_name = self.mdrun_keywords['s'][0].split('.')[0]
                except:
                    try:
                        tpr_name = self.mdrun_keywords['deffnm'][0].split('.')[0]
                    except:
                        tpr_name = 'log'

            END_CMD_GROMACS = f' >> {tpr_name}.lis 2>&1'

            for key in self.mdrun_keywords:
                if type(self.mdrun_keywords[key]) == bool:
                    CMD_GROMACS += f' -{key}'
                else:
                    CMD_GROMACS += f' -{key} {self.mdrun_keywords[key]}'
            job_str += CMD_GROMACS + END_CMD_GROMACS

        return job_str

    def write(self, name = 'job.sh'):
        with open(name, "w") as out:
            out.write(self.string())

if __name__ == '__main__':
    
    pass
    G = JOB(mdrun_keywords={'deffnm': 'mine'}, build_GROMACS_section='gmx mdrun -nt 12 -cpi -stepout 5000 -v -deffnm taka >& tak.lis')
    print(G.string())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re, os, tempfile, tqdm
from palmiche.utils import tools, xvg
import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
# I will deffine my cero potential as the potential of the system when the ligand is in the bulk without any interaction (the last poitn in the PMF)

def cmdwham(it = "tpr_files.dat",
            if_ = "pullf_files.dat",
            is_ = "coordsel.dat",
            unit = "kJ",
            out = 'data.xvg'):
    if os.path.exists(is_):
        tools.run(f"export GMX_MAXBACKUP=-1;gmx wham -it {it} -if {if_} -unit {unit} -is {is_} -o {out} -nice 19")
    else:
        tools.run(f"export GMX_MAXBACKUP=-1;gmx wham -it {it} -if {if_} -unit {unit} -o {out}-nice 19")


def pmf_5ave(path, pattern = 'coord[1-5]_selected.dat', plot = False, wham = False, **keywords):
    if wham: tmpout = tempfile.NamedTemporaryFile(suffix='.xvg')
    cwd = os.getcwdb()
    path = os.path.abspath(path)
    pattern = re.compile(pattern)
    reaction_coord = []
    pmf = []
    os.chdir(path)
    print("Generating PMF:")
    for item in tqdm.tqdm([i for i in tools.list_if_file(path) if pattern.match(i)]):
        if wham:

            cmdwham(is_ = item, out = tmpout.name)
            data = xvg.XVG(tmpout.name).data
        else:
            data = xvg.XVG(item).data
        reaction_coord.append(data[:,0])
        pmf.append(data[:,1])
  

    data = np.asarray([
                    np.mean(reaction_coord, axis = 0),
                    stats.sem(reaction_coord, axis = 0, ddof=1),
                    np.mean(pmf, axis = 0),
                    stats.sem(pmf, axis = 0, ddof=1)
                    ]).transpose()
    #data = np.random.randint(0,30,size = data.shape)
    if plot:
        X = data[:,0]
        Xerror = data[:,1]
        Y = data[:,2]
        Yerror = data[:,3]
        fig, ax = plt.subplots(**keywords)

        ax.errorbar(X, Y,xerr = Xerror)
        ax.fill_between(X, Y - Yerror, Y + Yerror, alpha = 0.2)

        ax.set(title = f"Mean PMF using the five channels [Kj/mol]",
        xlabel = r"$\xi$",
        ylabel = "PMF [kJ/mol]")
        plt.show()
        fig.savefig("PMF_5ave.svg")
    os.chdir(cwd)
    return data

def main(start_with = 'umbrella', pattern = 'coord[1-5]_selected.dat', wham = False, **keywords):
    cwd = os.getcwd()
    paths = sorted([os.path.abspath(os.path.join(p, "7e27", "sys_MMV007839_Cell_891_SP_param", "windows")) for p in tools.list_if_dir(".") if p.startswith(start_with)])
    string = f"{'simulation':<100}{'Y_error_mean':<10}\n"
    fig, ax = plt.subplots(**keywords)
    
    for path in paths:
        print(path)
        data = pmf_5ave(path, pattern, plot = False, wham = wham)
        for s in path.split('/'):
            if s.startswith(start_with):
                label = s.split(start_with)[-1].strip()
        X = data[:,0]
        Xerror = data[:,1]
        Y = data[:,2]
        Yerror = data[:,3]

        ope = np.round(np.mean(Yerror),2)
        string += f"{label:<100}{ope:<10}\n"

        ax.plot(X, Y, label=label)
        ax.fill_between(X, Y - Yerror, Y + Yerror, alpha = 0.2)
        
    os.chdir(cwd)
    print(string)
    
    ax.legend(loc="lower right")
    ax.set(title = "Different PMF parameters. Using Average.",
    xlabel = r"$\xi$",
    ylabel = "PMF [kJ/mol]")
    fig.savefig("PMF_resume_5ave.svg")
    fig.show()
    
    with open("PMF_error_5ave.txt", "w") as f:
        f.write(string)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import shutil, os
import re

def backoff(file_path):
    """
    

    Parameters
    ----------
    file : TYPE string
        DESCRIPTION: The name or the path f     or the specific file

    Returns
    -------
    None.
    If the file already exist. it will made a back up to ./#{file}.{str(i)}#,
    Where i is an integer.
    """
    basname = os.path.basename(file_path)
    dirname = os.path.dirname(file_path)
    if os.path.exists(file_path):
        new_basname = basname
        i = 1
        while(os.path.exists(os.path.join(dirname, new_basname))):
            new_basname = f"./#{basname}.{str(i)}#"
            i += 1
        print(f"Back Off! I just backed up {file_path} to {os.path.join(dirname, new_basname)}")
        shutil.copy2(file_path, os.path.join(dirname, new_basname))

class XVG:
    def __init__(self, file):
        self.file = file
        self.title = ""
        self.type = ""
        self.legend = []
        self.xlable = ""
        self.ylable = ""
        self.data = []
        self.multidata = []
        self.parser()

    def parser(self):
        with open(self.file, "r") as f:
            for line in f.readlines():
                if line.startswith("#"):
                    continue
                if line.startswith("@"):
                    if " title" in line:
                        self.title = line.split("title")[1].strip().replace("\"", "")
                    elif "TYPE" in line:
                        self.type = line.split()[1]
                    elif "xaxis  label" in line:
                        self.xlabel = line.split("xaxis  label")[1].strip().replace("\"", "")
                    elif "yaxis  label" in line:
                        self.ylabel = line.split("yaxis  label")[1].strip().replace("\"", "")
                    # Looking for the labels
                    item = re.search('@ s[0-9]+ legend', line) # Match @ s25 legend
                    if item:
                        self.legend.append(re.search("\"([^;]*)\"", line).group(1)) # Match the name of the label
                    continue
                if line.startswith("&"):
                    self.data = np.asarray(self.data, dtype=float)
                    self.multidata.append(self.data)
                    self.data = []
                    continue

                self.data.append(line.split())
        self.data = np.asarray(self.data, dtype=float)

    def write(self, out_file = None):
        if out_file:
            pd.DataFrame(self.data).to_string(out_file, index = False, header = False, float_format= '%.5f')
        else:
            backoff(self.file)
            pd.DataFrame(self.data).to_string(self.file, index = False, header = False, float_format= '%.5f')

    def quickgraph(self):
        fig, ax = plt.subplots()
        NUM_COLORS = self.data.shape[1]
        cm = plt.get_cmap('gist_rainbow')
        ax.set_prop_cycle('color', [cm(1.*i/NUM_COLORS) for i in range(NUM_COLORS)])
        for i in range(1, self.data.shape[1]):
            ax.plot(self.data[:,0], self.data[:,i], label = (self.legend[i-1]))
        ax.set(
            title = self.title,
            xlabel = self.xlabel,
            ylabel = self.ylabel,
        )
        fig.legend()
        plt.show()
if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
                        help = "xvg file",
                        dest='xvg',
                        type=str)
    args = parser.parse_args()
    xvg = XVG(args.xvg)
    #print(xvg.xlabel, xvg.title, xvg.ylabel)
    xvg.quickgraph()

                    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import tempfile, os, tqdm, pickle, json
from openbabel import openbabel as ob
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Descriptors
from rdkit.Chem import SaltRemover
from rdkit.Chem import rdmolops
from rdkit.Chem import rdDepictor
from rdkit.Chem.Draw import rdMolDraw2D
import pubchempy as pcp
import multiprocessing as mp
from datetime import datetime

 
# Consider the use of io.StringIO, in cases that 


def update_compounds(references, Threshold = 80, compounds_file = 'compounds.pickle', checkpoint_file = "cids_checkpoint.pickle"):
    # data = {"Last_update": now, "References": references, "Compounds": new_cids}
    now = datetime.now()#.strftime("%Y-%m-%d %H:%M:%S")

    print(f'Similarity search. Getting CIDs:...')
    uniques = []
    for ref in tqdm.tqdm(references):
        uniques += pcp.get_cids(ref[0] ,ref[1], searchtype='similarity', Threshold = Threshold)
    uniques = set(uniques)

    # Searching for non download entrance and adding to the queue
    if os.path.isfile(checkpoint_file):
        with open(checkpoint_file, 'rb') as f:
            uniques.update(pickle.load(f)['CID'])

    # Searching for the CID already on the database
    if os.path.isfile(compounds_file):
        with open(compounds_file, 'rb') as f:
            uniques -= set(c.cid for c in pickle.load(f)['Compounds']) # Only those CID not in the data base

    # Here I could restrict to only some part of that object if the data is to big

    print(f'Updating compounds data base {compounds_file} ...')
    new_compounds = []
    non_downloaded_cids = []
    for cid in tqdm.tqdm(list(uniques)):
        # The try except block is because the possible reject of the server
        # And in that case we need to save the data.
        try:
            new_compounds.append(pcp.Compound.from_cid(cid))
        except:
            non_downloaded_cids.append(cid)
            next
    
    
    print(f"The data base {compounds_file} was successfully updated!")
    # 
    if non_downloaded_cids:
        CIDs = {"Last_update": now, "CID": non_downloaded_cids}
        with open(checkpoint_file, 'wb') as f:
            pickle.dump(CIDs, f)
    else:# All CIDs were downloaded
        if os.path.isfile(checkpoint_file):
            os.remove(checkpoint_file)
            print(f"The {checkpoint_file} is not longer needed, therefore was removed")

    # Opening the compounds in case that exists

    if os.path.isfile(compounds_file):
        with open(compounds_file, 'rb') as f:
            data = pickle.load(f)
            # Updating
            
            # Datetime
            data["Last_update"] = now
            # References
            for ref in references:
                if ref not in data["References"]:
                    data["References"].append(ref)
            # Compounds
            data["Compounds"] += new_compounds

    else:
        # Initialization of the data
        data = {"Last_update": now, "References": references, "Compounds": new_compounds}
    
    # Writting
    with open(compounds_file, 'wb') as f:          
        pickle.dump(data, f)


    return new_compounds


def obconvert(inpath, outpath):
    """Convert  molecule ussing openbabel

    Args:
        input (str, path): input molecule.
        output (str, path): must have the extention of the molecule.
    """
    in_ext = os.path.basename(inpath).split('.')[-1]
    out_ext = os.path.basename(outpath).split('.')[-1]

    obConversion = ob.OBConversion()
    obConversion.SetInAndOutFormats(in_ext, out_ext)
    mol = ob.OBMol()
    obConversion.ReadFile(mol, inpath)   # Open Babel will uncompressed automatically
    #mol.AddHydrogens()
    obConversion.WriteFile(mol, outpath)

def confgen(smiles, outpath = "mol.pdbqt"):
    """Create a 3D model from smile to

    Args:
        smiles (str): a valid smiles code.
        outpath (str, optional): The output molecule with its corresponding extension. Defaults to "mol.pdbqt".
    """
    moltemp = tempfile.NamedTemporaryFile(suffix='.mol')
    mol = Chem.AddHs(Chem.MolFromSmiles(smiles))
    AllChem.EmbedMolecule(mol)
    AllChem.MMFFOptimizeMolecule(mol)
    Chem.MolToMolFile(mol, moltemp.name)
    obconvert(moltemp.name, outpath)
    with open(outpath, 'r') as o:
        string = o.read()
    return string



# def lipinski_filter_pcp(compound_obj, maxviolation = 2):
#     cont = 0
#     if compound_obj.h_bond_donor_count > 5: cont += 1
#     if compound_obj.h_bond_acceptor_count > 10: cont += 1
#     if compound_obj.exact_mass > 500: cont += 1
#     if compound_obj.xlogp > 5: cont += 1
#     #if compound_obj.rotatable_bond_count > 10: cont += 1
#     if cont >= maxviolation: return False
#     return True

# def elements_filter_pcp(compound_obj):
#     vina_elements = ['Br']
#     for atom in compound_obj.atoms:
#         if atom.element not in vina_elements:
#             return False
#     return True

# def elements_filter_pcp(mol):
#     vina_elements = ['Br']
#     for atom in mol.GetAtoms():
#         if atom.GetSymbol() not in vina_elements:
#             return False
#     return True

def lipinski_filter(mol, maxviolation = 2):
    filter = {
        'HBD': {'method':Descriptors.rdMolDescriptors.CalcNumLipinskiHBD,'cutoff':5},
        'HBA': {'method':Descriptors.rdMolDescriptors.CalcNumLipinskiHBA,'cutoff':10}, 
        'wt': {'method':Descriptors.MolWt,'cutoff':500},
        'MLogP': {'method':Descriptors.MolLogP,'cutoff':5}
        #'rotbond': {'method': Descriptors.NumRotatableBonds(), 'cutoff':10}
    }
    cont = 0
    for property in filter:
        
        if filter[property]['method'](mol) > filter[property]['cutoff']:
            cont += 1
        if cont >= maxviolation:
            return False
    return True

def remove_salt(mol):
    remover = SaltRemover.SaltRemover()
    return remover.StripMol(mol, dontRemoveEverything=True)

def neutralize_atoms(mol):
    """Neutralizing Molecules¶
    Author: Noel O’Boyle (Vincent Scalfani adapted code for RDKit)  
    Source: https://baoilleach.blogspot.com/2019/12/no-charge-simple-approach-to.html
    Index ID#: RDKitCB_33
    Summary: Neutralize charged molecules by atom.
    https://www.rdkit.org/docs/Cookbook.html?highlight=neutralize

    Args:
        mol (rdkit object): A rdkit molecules

    Returns:
        rdkit molecules: the molecule after neutralization
    """
    pattern = Chem.MolFromSmarts("[+1!h0!$([*]~[-1,-2,-3,-4]),-1!$([*]~[+1,+2,+3,+4])]")
    at_matches = mol.GetSubstructMatches(pattern)
    at_matches_list = [y[0] for y in at_matches]
    if len(at_matches_list) > 0:
        for at_idx in at_matches_list:
            atom = mol.GetAtomWithIdx(at_idx)
            chg = atom.GetFormalCharge()
            hcount = atom.GetTotalNumHs()
            atom.SetFormalCharge(0)
            atom.SetNumExplicitHs(hcount - chg)
            atom.UpdatePropertyCache()
    return mol

def largest_mol(mol):
    """
    Author: Andrew Dalke and Susan Leung
    Source: https://sourceforge.net/p/rdkit/mailman/message/36355644/ and https://github.com/susanhleung/rdkit/blob/dev/GSOC2018_MolVS_Integration/rdkit/Chem/MolStandardize/tutorial/MolStandardize.ipynb
    Index ID#: RDKitCB_31
    Summary: Select largest fragment from a molecule
    """
    mol_frags = rdmolops.GetMolFrags(mol, asMols = True)
    if len(mol_frags) == 1:
        return mol
    return max(mol_frags, default=mol, key=lambda m: m.GetNumAtoms())
    
    #largest_Fragment = rdMolStandardize.LargestFragmentChooser()
    #return largest_Fragment.choose(mol)


def compounds_filter(pcp_compound):
    pdbqt_tmp = tempfile.NamedTemporaryFile(suffix='.pdbqt')

    smiles = pcp_compound.isomeric_smiles
    compound = {'CID': pcp_compound.cid, 'isomeric_smiles_pcp': smiles}
    mol = Chem.MolFromSmiles(smiles)
    mol = remove_salt(mol)
    
    #un = rdMolStandardize.Uncharger()
    #un.uncharge(mol)
    mol = neutralize_atoms(mol)
    mol = largest_mol(mol)
    compound['rdkit_smiles'] = Chem.MolToSmiles(mol)
    if lipinski_filter(mol):
        pdbqt_string = confgen(compound['rdkit_smiles'], outpath = pdbqt_tmp.name)
        compound['pdbqt_string'] = pdbqt_string
        return compound
    else:
        return None

def receptor(pdbqt_file, box_file, receptor_name):
    with open(pdbqt_file, 'r') as f:
        pdbqt_string = f.read()
    with open(box_file, 'r') as f:
        grid_opt = json.load(f)[receptor_name]
    return {'name': receptor_name, 'pdbqt_string': pdbqt_string, 'grid_opt': grid_opt}

def ligands(compounds, cpus = 0):

    if cpus:
        jobs = cpus
    else:
        jobs = mp.cpu_count()
    pool = mp.Pool(jobs)

    data =  list(filter(None, pool.imap_unordered(compounds_filter, compounds)))
    pool.close()
    
    dictionary = dict()
    for lig in data:
        dictionary[lig['CID']] = lig
    return dictionary

def vinaconfig(receptor_path,
                ligand_path,
                grid_center,
                size,
                num_modes = 1,
                exhaustiveness = 8,
                cpu = 3):

    name = os.path.basename(ligand_path).split(".")[0]
   
    string =f"{'receptor':25} = {receptor_path}\n"\
            f"{'ligand':25} = {ligand_path}\n\n"\
            f"{'center_x':25} = {grid_center[0]}\n"\
            f"{'center_y':25} = {grid_center[1]}\n"\
            f"{'center_z':25} = {grid_center[2]}\n\n"\
            f"{'size_x':25} = {size[0]}\n"\
            f"{'size_y':25} = {size[1]}\n"\
            f"{'size_z':25} = {size[2]}\n\n"\
            f"{'num_modes':25} = {num_modes}\n"\
            f"{'exhaustiveness':25} = {exhaustiveness}\n"\
            f"{'cpu':25} = {cpu}\n\n"\
            f"{'out'} =  {name}_out.pdbqt\n"
    return string
            
def moltosvg(mol, molSize = (300,300), kekulize = True):
    mc = Chem.Mol(mol.ToBinary())
    if kekulize:
        try:
            Chem.Kekulize(mc)
        except:
            mc = Chem.Mol(mol.ToBinary())
    if not mc.GetNumConformers():
        rdDepictor.Compute2DCoords(mc)
    drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0],molSize[1])
    drawer.DrawMolecule(mc)
    drawer.FinishDrawing()
    svg = drawer.GetDrawingText()
    return svg #le quite lo de replace svg: por '', pq sino no funcionaba


    
    
    

# references = [
#     ('OC(CC(O)C(F)(F)C(F)(F)F)C1CCCNC1','smiles', 'cis1_bh267_m'),
#     ('COC1CC[C@@H](C(O)CC(O)C(F)(F)C(F)(F)F)C(O)C1','smiles', 'sys_MMV007839')
# ]
# AQdicionar update info al opbjecto, la fecha, las estructuras que se emplearon, el threhold, etc..

# # Checking for necessary updates
# new_cids = update_cids(references)
# new_compounds = update_compounds(new_cids)
# for compound in new_compounds:
#     # here I need to do a lot of stuffs:
#     # Sanitze the SMILES, removes sals and so,
#     # Generate a confomation
#     # Generate a pdbqt file
#     # generate the configuation file for vina
#     # run vina
#     # save the data inside the object

#     setattr(compound, 'docking', 'a dict with final SMILES used, the config file, the exhautivines used, the initial configuration files and the BE confomation, also the energy')


"""
Puedo poner un CID que tenga los CID aceptados y los rechazados, en fin, los cid que se
han anmalizado hasta la fecha, pero se le puede a;adir ese plus,
Esto te ayuda a que si los CID que estas proponiendo ya se estudiaron ni molestarse
Tengoq u obtener los SMILES de las dos moleculas que estnaos estudiando para que sean las referencias
la cosa es que tengo que hacer una funcion que me actualice los smiles,
segun las moleculas que se adicionene
la cosa es que voy a tener dos archivos, uno
.smi
que va a tener todas los smiles procesados, y como nombre el CID
Y otro archivo .sdf con los conformeros
FInalmente una carpeta con los .pdbqt que van a tener por nombre el CID
el update va a buscar el CID en smile, tambien puede buscar por smile,
En caso de no estar, delo a;adira a los res niveles.

HE ARPENDIDO UNA PILA DE TALLAS

1- Buscar por similitud en pubchem con pubchempy
lo que hacemos es descargar solo los CID porque sino al descargar tanta informacion el servidos nos va a rechazar de
por tanto:
a = pcp.get_cids('COC1=CC2=C(C=C1)C(=O)CC(O2)(C(C(F)(F)F)(F)F)O','smiles', searchtype='similarity', Threshold = 80)
a = pcp.get_cids(2917555 ,'cid', searchtype='similarity', Threshold = 80)
todos estos metodos funcionas que:
especificas el tipo a buscar ya sea smile, cid aid o lo que quieras,
pero tinees que poner el tipo que buscaste, por
luego especificas searchtype
https://pubchemdocs.ncbi.nlm.nih.gov/pug-rest$_Toc494865580
En esa pagina explican como se manja estos parametros
el threshold es para el tanimoto
ahora bien: generamos un set de varios cids
y luego llamamos a 


"""
if __name__ == '__main__':
    pass
    v = vinaconfig('./k', 'l', (5,8,9), (50,50,50))
    print(v)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from palmiche.utils import tools
import os, glob
import pandas as pd
import numpy as np
import statistics as stat
import matplotlib.pyplot as plt
import multiprocessing as mp
import tempfile


def COM_dist_chunk_code(conf_file,grouplist,ndx,tpr,prefix):
    """
    This function doesn't have any relevance. It is only a code chunk 
    in order to parallelize the distance calculation. This must be outside of the function

    """
    tmp_row = [int(os.path.basename(conf_file).split(prefix)[-1].split('.')[0])]#Conf_ID, in this part is where the number is
    # HERE I NEED TO PUT THE TEMPORAL FILES

    opt_tmp = tempfile.NamedTemporaryFile(suffix='.opt')
    xvg_tmp = tempfile.NamedTemporaryFile(suffix='_dist.xvg') 
    with open(opt_tmp.name, 'w') as f:
        for (i, groups) in enumerate(grouplist):
            f.write(f"\"{i}\" com of group {groups[0]} plus com of group {groups[1]};\n")
    tools.run(f"gmx distance -s {tpr} -f {conf_file} -n {ndx} -sf {opt_tmp.name} -oall {xvg_tmp.name} -xvg none")
    tmp_row += list(np.loadtxt(xvg_tmp.name)[1:])
        
    Mean = stat.mean(tmp_row[1:])
    try:
        StDev = stat.stdev(tmp_row[1:])
    except:
        StDev = 0.0
    tmp_row += [Mean, StDev]
        
    for i in range(1,len(tmp_row)):
        tmp_row[i] = round(tmp_row[i],3)
    return tmp_row


def main(grouplist, cpu = 0, ndx = 'index.ndx', tpr = 'pull.tpr', xtc = 'pull.xtc', prefix = 'conf', out = 'summary_distances.dat', split_out_dir = '.'):
    """
    

    Parameters
    ----------
    grouplist : TYPE
        DESCRIPTION. A list of list,
        e.g.:
                ligands = ['LIA', 'LIB', 'LIC', 'LID', 'LIE']
                grouplist = [[ligand,ligand+"_CLOSE_AA"] for ligand in ligands]
                so...
                [(group1,group2), (group3,group4),...]
    cpu : TYPE, optional int
        DESCRIPTION. The default is 0. It will launch  as much as cpu has 
        the node the function COM_dist_chunk_code(). If some number is specified
        this will be the launched time of the function COM_dist_chunk_code().
    ndx : TYPE, optional
        DESCRIPTION. The default is 'index.ndx'.
    tpr : TYPE, optional
        DESCRIPTION. The default is 'pull.tpr'.
    xtc : TYPE, optional
        DESCRIPTION. The default is 'pull.xtc'.
    prefix : TYPE, optional
        DESCRIPTION. The default is 'conf'.
    out : TYPE, optional
        DESCRIPTION. The default is 'summary_distances.dat'.

    Returns
    -------
    None.

    """
    
    if not os.path.exists(ndx): 
        raise FileNotFoundError(f"\"{ndx}\" doesn't exist or is not accessible.")
    if not os.path.exists(tpr): 
        raise FileNotFoundError(f"\"{tpr}\" doesn't exist or is not accessible..")
    if not os.path.exists(xtc): 
        raise FileNotFoundError(f"\"{xtc}\" doesn't exist or is not accessible..")
    
    print('Splitting the trajectory...')
    tools.run(f"export GMX_MAXBACKUP=-1; echo 'system' | gmx trjconv -s {tpr} -f {xtc} -o {os.path.join(split_out_dir, prefix)}.gro -sep")
    print('Done!')
    conf_files = sorted(glob.glob(f"{os.path.join(split_out_dir, prefix)}[0-9]*"), key=lambda x:int(os.path.basename(x).split(prefix)[-1].split('.')[0]))
    
    columns = ["Conf_ID"] + [f"{item[0]}-{item[1]}" for item in grouplist] + ["Mean","StDev"]        
    data = []

    #In order to parallelize the dist calculation
    if cpu:
        jobs = cpu
    else:
        jobs = mp.cpu_count()
    pool = mp.Pool(jobs)
    
    data = pool.starmap(COM_dist_chunk_code, [(conf_file,grouplist,ndx,tpr,prefix) for conf_file in conf_files])
    
    pool.close()
    
    general_table = pd.DataFrame(data=data, columns = columns)
    general_table.plot(x="Conf_ID", y=columns[1:-1])
    plt.xlabel("Conf_ID")
    plt.ylabel("COM distance (nm)")
    plt.savefig('summary_distances.svg')
    #plt.show()
    
    
    with open(out, "w") as f:
        general_table.to_string(f,index=False)
    #general_table.to_csv(out, sep='\t', index=False)
    with open("mean.dat", "w") as f:
        general_table[["Conf_ID", "Mean"]].to_string(f,index=False, header=False)
    
    
if __name__=='__main__':
    grouplist = []
    for char in 'ABCDE':
        grouplist.append([f"LI{char}", f"LI{char}_CLOSE_AA"])
    path ="/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param"
    tools.makedirs('split_xtc')
    main(grouplist, cpu=6, ndx= os.path.join(path, 'index.ndx'), tpr=os.path.join(path,'pull.tpr'),
                    xtc=os.path.join(path, 'pull.xtc'), prefix='conf', out='summary_distances.dat', split_out_dir='split_xtc')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from palmiche.utils import xvg, potentials
import matplotlib.pyplot as plt
import numpy as np
orient_restrain_init = 30
pullf = ("pull_pullf.xvg")
pullx = ("pull_pullx.xvg")
# Here I will get the info of the COM and the orientation restrain
t = xvg.XVG(pullf).data[:,0]
plt.subplot(2, 1, 1)
plt.ylabel("V [kJ/mol]")
for coord in range(11, xvg.XVG(pullf).data.shape[1]):
    f = np.pi/180*xvg.XVG(pullf).data[:,coord]
    x = xvg.XVG(pullx).data[:,coord]    
    pot = potentials.POT("flat_bottom", f, x, ref = orient_restrain_init).pot
    plt.plot(t, pot, label = str(coord))
    #plt.ylim(0,2000)
plt.legend(loc="upper right")

plt.subplot(2, 1, 2)
plt.xlabel("Time")
plt.ylabel("COM [deg]")
for coord in range(11, xvg.XVG(pullf).data.shape[1]):
    x = xvg.XVG(pullx).data[:,coord]    
    plt.plot(t, x, label = str(coord))
    #plt.ylim(0,60)
plt.legend(loc="upper right")   
plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# We will use semantic version (major, minor, patch)
__main_version_tuple__ = (0, 0, 0)
__pre_version_tuple__ = ('alpha',4) # or None or an empty tuple, the first char is alpha, beta, rc, etc...

if __pre_version_tuple__:
    __version_tuple__ = tuple(list(__main_version_tuple__) + list(__pre_version_tuple__))
    __version__ = '.'.join([str(i) for i in __main_version_tuple__]) + f'-{__pre_version_tuple__[0]}' + '.'.join([str(i) for i in __pre_version_tuple__[1:]])
else:
     __version__ = '.'.join([str(i) for i in __main_version_tuple__])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from palmiche.utils import tools, pdb, gro
import os

                
                    
def union(list_of_list):
    """
    

    Parameters
    ----------
    list_of_list : TYPE, list
        DESCRIPTION. like
        [[1,2,1,3,4,5],[5,5,6,6,4,1,8,1],[1,1,10,5,20,36],[55,5,1,45,1]]


    Returns
    -------
    TYPE
        a list with the elemnts in comun between all the lists

    """
    to_return = list_of_list[0]
    
    for item in list_of_list:
        to_return = set(to_return) & set(item)  
    #    to_return = list(filter(lambda x: x in to_return, item))
    return sorted(to_return)
    #return sorted(list(set(to_return)))

def ndx_rectification(index_file, config_path, grouplist):
    """
    This function is for the case that you have more than one monomer fo the 
    same protein. In this case each monomer should have the same residues (number)
    This function looks for the common residues of the groups decleared in 
    grouplist and then look for the index atom of these aminoacids 
    on each monomer. Then it will try to guess to which group belong the new
    index and finally export a dictionary that could be write to a GROMACS
    index file using the function the method write of the class NDX.
    The new atom index included all the atoms in the residues (also Hydrogens)

    Parameters
    ----------
    index_file : TYPE
        DESCRIPTION.
    config_path : TYPE
        DESCRIPTION.
    grouplist : TYPE
        DESCRIPTION.

    Returns
    -------
    new_index_dict : TYPE
        DESCRIPTION.

    """
    old_index_dict = tools.NDX(index_file).data
    new_index_dict = old_index_dict.copy()
    basename = os.path.basename(config_path)
    ext = basename.split('.')[-1]
    to_work = []
    
    if ext == 'gro':
        atoms = gro.GRO(config_path).atoms
        
        for group in grouplist:
            residues = []
            for old_index in old_index_dict[group]:
                residues.append(atoms[old_index-1].resNumber)
            to_work.append(list(set(residues)))

        common_resi = union([i for i in to_work])
        print("These are the aa taken into account, use the following command to see them in Pymol:\n show sticks, resi "+"+".join([str(i) for i in common_resi]))
        
        
        test = atoms[0].resNumber
        tmp_index = []
        total_index = []
        for i, atom in enumerate(atoms):
            if atom.resNumber < test or i == len(atoms)-1:
                    
                total_index.append(tmp_index)
                tmp_index = []                
            if atom.resNumber in common_resi:
                tmp_index.append(i+1)
            test = atom.resNumber

        #Guesing the groups
        for group in grouplist:
            old_index = new_index_dict[group]
            intersection = 0

            for new_index in total_index:
                tmp = len(set(old_index) & set(new_index))
                
                if tmp > intersection:
                    intersection = tmp
                    new_index_dict[group] = new_index
            print(f"The new index group of {group} has {intersection} atoms that also were in the old index group that had a total of {len(old_index)} atoms")
        
        return new_index_dict
    
    elif ext == 'pdb':
        atoms = pdb.PDB(config_path).atoms
        
        for group in grouplist:
            residues = []
            for old_index in old_index_dict[group]:
                residues.append(atoms[old_index-1].resSeq)
            to_work.append(list(set(residues)))

        common_resi = union([i for i in to_work])
        print("These are the aa taken into account, use the following command to see them in Pymol:\n show sticks, resi "+"+".join([str(i) for i in common_resi]))
        
        
        test = atoms[0].resSeq
        tmp_index = []
        total_index = []
        for i, atom in enumerate(atoms):
            if atom.resSeq < test or i == len(atoms)-1:
                    
                total_index.append(tmp_index)
                tmp_index = []                
            if atom.resSeq in common_resi:
                tmp_index.append(i+1)
            test = atom.resSeq

        #Guesing the groups
        for group in grouplist:
            old_index = new_index_dict[group]
            intersection = 0

            for new_index in total_index:
                tmp = len(set(old_index) & set(new_index))
                
                if tmp > intersection:
                    intersection = tmp
                    new_index_dict[group] = new_index
            print(f"The new index group of {group} has {intersection} atoms that also where in the old index group that had a total of {len(old_index)} atoms")
        
        return new_index_dict
    
    else:
        return(f"The file {config_path} has not a gro or pdb extension")

                  

def flat_bottom_itp_maker(file_path, geometry = 8, r = 'FLAT_BOTOM_r', k = 'FLAT_BOTOM_k', output = 'flat_bottom_posre.itp', H_atoms = False, backup = True):
    """
    

    Parameters
    ----------
    file_path : TYPE path to a gro pdb or itp file
        The configuration fiel from where .
    geometry : TYPE (int), optional
        DESCRIPTION. The default is 8.
    r : TYPE (float)
        DESCRIPTION. The default is 'FLAT_BOTOM_k'. And this flat is controlled with the mdp. You could set a number also, 1.0
    output : TYPE (string), optional
    k : TYPE (str or float), optional
        DESCRIPTION. The default is 'FLAT_BOTOM_k'. And this flat is controlled with the mdp. You could set a number also, 600
    output : TYPE (string), optional
        DESCRIPTION. The default is 'flat_bottom_posre.itp'.
    H_atoms : TYPE (bool), optional
        DESCRIPTION. The default is False. Control the flag of the function
        tools.get_atom_index()
    
    For the description of geometry, r and k, please consult the GROMACS manual:
        https://manual.gromacs.org/
    Returns
    -------
    The posre file with the flat bottom potential option
    if and itp file was given, it will add at the end of the itp the
    corresponding reference lines for the posre file.

    """
    func_type = 2 #This is for flat bottom potential in GROMACS
    if backup: tools.backoff(output)
    atom_index = tools.get_atom_index(file_path, H_atoms=H_atoms)
    
    basename = os.path.basename(file_path)
    ext = basename.split('.')[-1].lower()
    to_print = [f" {index:<15}{func_type:<15}{geometry:<15}{r:<15}{k:<15}\n" for index in atom_index]
    with open(output,'w') as out:
        out.write(";Generated by make_posre.py\n"\
                 f";{'index':<15}{'func_type':<15}{'geometry':<15}{'r':<15}{'k':<15}\n")
        for item in to_print:
            out.write(item)
    if ext == 'itp':
        if backup: tools.backoff(file_path)
        with open(file_path, 'a') as f:
            f.write("\n; Include Position restraint file for Flat Bottom Potential\n"\
                    "#ifdef POSRES\n"\
                    f"#include \"{output}\"\n"\
                    "#endif\n")
    
def flat_bottom_config_posre(config_path,resNameList, H_atoms= False, backup = True):
    """
    !!!!!Tengo que mejorar esta funcion, creo que es mejor trabajr con el index file
    tanto para este como para  el itp generator, porque de este modo, se puede ser mas general
    y la seleccion no se restringe a solo una molecula
    Ademas, esto dara problemas en el caso que se le aplique a una seccion de la proteina
    y como mi caso, sea un penimport gro_reader, pdb_readertamero,
    pues aplicara el potencial en todos los monomeros,
    debido a que los aa se llaman iguales,
    y tal bez no es lo que se desee!!!!!!

    Parameters
    ----------
    config : TYPE path 
        DESCRIPTION. configuration file (pdb or gro)
    resNameList : TYPE list
        DESCRIPTION. List of Residues names for which the flatt
        bottom will be used
        ['LIA', 'LIB', 'LIC', ...]
    H_atoms : TYPE bool
        DESCRIPTION. The default is False. Control the flag of the function
            mol_selection_center(), 

        

    Returns
    -------
    The function will take the center for each residues in resNameList
    And then the xyz values of                 else:
                    select.append(atom)the instance will be changes for the selection center
    After this, the config file will be updated, and the original will be backup

    """
    basename = os.path.basename(config_path)
    ext = basename.split('.')[-1]
    if ext == 'gro':
        config_data = gro.GRO(config_path)
        for resName in resNameList:
            xc, yc, zc = config_data.centroid(resName_selection = [resName], H_atoms=H_atoms)
            for atom in config_data.atoms:
                if atom.resName == resName:
                    if H_atoms:
                        atom.x = xc
                        atom.y = yc
                        atom.z = zc
                    else:
                        if atom.atomName[0] not in 'Hh':
                            atom.x = xc
                            atom.y = yc
                            atom.z = zc                        
                
    elif ext == 'pdb':
        config_data = pdb.PDB(config_path)
        for resName in resNameList:
            xc, yc, zc = config_data.centroid(resName_selection = [resName], H_atoms=H_atoms)
            for atom in config_data.atoms:
                if atom.resName == resName:
                    if H_atoms:
                        atom.x = xc
                        atom.y = yc
                        atom.z = zc                       
                    else:
                        if atom.name[0] not in 'Hh':
                            atom.x = xc
                            atom.y = yc
                            atom.z = zc
    else:
        return('The extension file is not pdb or gro')
    config_data.write(backup = backup)
        


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from palmiche.utils import xvg, tools
import matplotlib.pyplot as plt
import os
import numpy as np

# I will deffine my cero potential as the potential of the system when the ligand is in the bulk without any interaction (the last poitn in the PMF)

def cmdwham(it = "tpr_files.dat",
            if_ = "pullf_files.dat",
            is_ = "coordsel.dat",
            unit = "kJ",
            nBootstrap = 100,
            bs_method = "b-hist",
            bsres = "bsResult.xvg",
            bsprof = "bsProfs.xvg"):
    if os.path.exists(is_):
        tools.run(f"gmx wham -it {it} -if {if_} -unit {unit} -is {is_} -nBootstrap {nBootstrap} -bs-method {bs_method} -bsres {bsres} -bsprof {bsprof} -nice 19")
    else:
        tools.run(f"gmx wham -it {it} -if {if_} -unit {unit} -nBootstrap {nBootstrap} -bs-method {bs_method} -bsres {bsres} -bsprof {bsprof} -nice 19")

def main(wham = True, nBootstrap = 100, figsize = (16,9), start_with = 'umbrella'):
    cwd = os.getcwd()
    paths = sorted([os.path.abspath(os.path.join(p, "7e27", "sys_MMV007839_Cell_891_SP_param", "windows")) for p in tools.list_if_dir(".") if p.startswith(start_with)])
    string = f"{'simulation':<100}{'Y_error_mean':<10}\n"
    fig, ax = plt.subplots(figsize = figsize)
    for path in paths:

        os.chdir(path)
        if wham:
            cmdwham(nBootstrap = nBootstrap, is_ = "coord0_selected.dat")
        data = xvg.XVG("bsResult.xvg").data
        for s in path.split('/'):
            if s.startswith(start_with):
                label = s.split(start_with)[-1].strip()
        X = data[:,0]
        Y = data[:,1] - data[-1:,1]
        Yerror = data[:,2]
        ope = np.round(np.mean(Yerror),2)
        string += f"{label:<100}{ope:<10}\n"

        ax.plot(X, Y, label=label)
        ax.fill_between(X, Y - Yerror, Y + Yerror, alpha = 0.2)
        
    os.chdir(cwd)
    print(string)
    
    ax.legend(loc="lower right")
    ax.set(title = f"Different PMF parameters. nBootstrap = {nBootstrap}",
    xlabel = r"$\xi$",
    ylabel = "PMF [kJ/mol]")
    fig.savefig("PMF_resume.svg")
    fig.show()
    
    with open("PMF_error_mean.txt", "w") as f:
        f.write(string)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from palmiche.utils import tools
import numpy as np

class InputERROR(Exception):
    pass

class Atom:
    def __init__(self, line):
        #The last is not take
        self.resNumber = int(line[:5]) #residue number (5 positions, integer)
        self.resName = line[5:10].strip() #residue name (5 characters)
        self.atomName = line[10:15].strip() #atom name (5 characters)
        self.atomNumber = int(line[15:20]) #atom number (5 positions, integer)
        #position (in nm, x y z in 3 columns, each 8 positions with 3 decimal places)
        self.x = float(line[20:28]) 
        self.y = float(line[28:36])
        self.z = float(line[36:44])
        #velocity (in nm/ps (or km/s), x y z in 3 columns, each 8 positions with 4 decimal places)
        try: 
            self.vX = float(line[44:52])
        except:
            self.vX = 0.0000
        try: 
            self.vY = float(line[52:60])
        except:
            self.vY = 0.0000   
        try: 
            self.vZ = float(line[60:68])
        except:
            self.vZ = 0.0000
        
    def __getitem__(self, key):
        return self.__dict__[key]

class VECTOR():
    def __init__(self,line):
        split = line.split()
        try: 
            self.v1x = float(split[0])
        except: 
            self.v1x = 0.0
        try: 
            self.v2y = float(split[1])
        except: 
            self.v2y = 0.0 
        try: 
            self.v3z = float(split[2])
        except: 
            self.v3z = 0.0        
        try: 
            self.v1y = float(split[3])
        except: 
            self.v1y = 0.0        
        try: 
            self.v1z = float(split[4])
        except: 
            self.v1z = 0.0  
        try: 
            self.v2x = float(split[5])
        except: 
            self.v2x = 0.0      
        try: 
            self.v2z = float(split[6])
        except: 
            self.v2z = 0.0 
        try: 
            self.v3x = float(split[7])
        except: 
            self.v3x = 0.0  
        try: 
            self.v3y = float(split[8])
        except: 
            self.v3y = 0.0
        
    def __getitem__(self, key):
        return self.__dict__[key]
    def write(self):
        string = "%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f\n"%\
            (self.v1x,self.v2y,self.v3z,self.v1y,self.v1z,self.v2x,self.v2z
             ,self.v3x,self.v3y)
        return string
class Residue:
    """
    This will convert a list of Atom objects that belongs and will create a residue class taken into account the information of the first atom. It supoouse that all of them belong to the same residue of
    """
    def __init__(self, atoms):
        self.resName = atoms[0].resName
        self.resNumber = atoms[0].resNumber
        self.atoms = atoms

    def __getitem__(self, key):
        return self.__dict__[key]

class GRO:

    def __init__(self, file):
        self.file = file
        self.atoms = []
        self.name = ''
        self.NumberAtoms = 0
        self.vector = ""
        self.parse()

        
    def parse(self):
        with open(self.file, 'r') as f:
            lines = f.readlines()
        self.name = lines[0].strip()
        self.NumberAtoms = int(lines[1])
        self.vector = VECTOR(lines[-1])
        
        if len(lines) != self.NumberAtoms + 3: raise InputERROR('The number of atoms do not match or is something wrong with the gro file. See: https://manual.gromacs.org/archive/5.0.4/online/gro.html')
        
        for line in lines[2:-1]:
            atom = Atom(line)
            self.atoms.append(atom)
            
        
    def get_atoms(self, to_dict = True):
        """Return a list of all atoms.

        If to_dict is True, each atom is represented as a dictionary.
        Otherwise, a list of Atom objects is returned."""
        
        if to_dict: return [x.__dict__ for x in self.atoms]
        else: return self.atoms
        

    def get_residues(self):
        """Rearrange the data as residues function.

        Returns:
            list: a list of Residues objects
        """
        residues = []
        i = 0
        while i < len(self.atoms):
            resAtoms = []
            resName = self.atoms[i].resName
            resNumber = self.atoms[i].resNumber
            j = i
            while self.atoms[j].resName == resName and self.atoms[j].resNumber == resNumber:
                resAtoms.append(self.atoms[j])
                j += 1
                if j >= len(self.atoms): break
            residues.append(Residue(resAtoms))
            i = j
        
        return residues

    def edges(self):
        x_min = min(self.atoms, key=lambda item: item.x)
        y_min = min(self.atoms, key=lambda item: item.y)
        z_min = min(self.atoms, key=lambda item: item.z)

        x_max = max(self.atoms, key=lambda item: item.x)
        y_max = max(self.atoms, key=lambda item: item.y)
        z_max = max(self.atoms, key=lambda item: item.z)
        
        print(f"""
              X axes:
                     Length: {round(x_max.x-x_min.x, 3)} nm
                     Bounding atoms: {x_max.atomName}_{x_max.atomNumber} ({x_max.resName}) ----- {x_min.atomName}_{x_min.atomNumber} ({x_min.resName})
              Y axes:
                     Length: {round(y_max.y-y_min.y, 3)} nm
                     Bounding atoms: {y_max.atomName}_{y_max.atomNumber} ({y_max.resName}) ----- {y_min.atomName}_{y_min.atomNumber} ({y_min.resName})
              Z axes:
                     Length: {round(z_max.z-z_min.z, 3)} nm
                     Bounding atoms: {z_max.atomName}_{z_max.atomNumber} ({z_max.resName}) ----- {z_min.atomName}_{z_min.atomNumber} ({z_min.resName})
            """)
        
    def box_center(self, resName_selection = [], H_atoms = True):
        if resName_selection:
            atoms = [atom for atom in self.atoms if atom.resName in resName_selection]
        else:
            atoms = self.atoms
        
        if not H_atoms:
            atoms = [atom for atom in atoms if atom.atomName[0] not in 'Hh']

        x_min = min(atoms, key=lambda item: item.x)
        y_min = min(atoms, key=lambda item: item.y)
        z_min = min(atoms, key=lambda item: item.z)

        x_max = max(atoms, key=lambda item: item.x)
        y_max = max(atoms, key=lambda item: item.y)
        z_max = max(atoms, key=lambda item: item.z)
         
        return (round((x_max.x+x_min.x)/2,3), round((y_max.y+y_min.y)/2,3), round((z_max.z+z_min.z)/2,3))

    def centroid(self, resName_selection = [], H_atoms = True):
        
        if resName_selection:
            atoms = [atom for atom in self.atoms if atom.resName in resName_selection]
        else:
            atoms = self.atoms
        
        if not H_atoms:
            atoms = [atom for atom in atoms if atom.atomName[0] not in 'Hh']
        
        centroid = (sum([np.array([atom.x,atom.y,atom.z]) for atom in atoms]) / len(atoms))
        centroid = [round(coord, 3) for coord in centroid]
        return tuple(centroid)
    
    def write(self, out_file = None, backup = True):
        #This could be used if some modifications were made in the instance
        if out_file:
            to_work = out_file
        else:
            to_work = self.file
            if backup: tools.backoff(to_work)

        with open(to_work,'w') as out:
            if self.name: out.write(self.name+'\n')
            out.write(str(self.NumberAtoms)+'\n')
            for atom in self.atoms:
                out.write("%5i%5s%5s%5i%8.3f%8.3f%8.3f%8.4f%8.4f%8.4f\n"%
                          (atom.resNumber,atom.resName,atom.atomName, 
                          atom.atomNumber,atom.x,atom.y,atom.z,atom.vX,atom.vY,
                          atom.vZ
                          ))            
            if self.vector: out.write(self.vector.write())
            
            
if __name__ == '__main__':
    #import sys
    #gro = GRO(sys.argv[1])
    gro = GRO('confout.gro')
    atoms = gro.get_atoms(to_dict = False)
    gro.edges()
    print(gro.box_center())
    
    for atom in gro.atoms:
        if atom.resName == 'LIA':
            atom.x = 0
            atom.y = 0
            atom.z = 0

    
    gro.write()
    #print(gro.NumberAtoms)
    #print(gro.name)
   
    



 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from palmiche.utils import tools
from palmiche.utils import pdb
from palmiche.utils.pKa2pdb2gmx import pka2gmx
from glob import glob
import tempfile



def get_path_dict(ligands_path = "./ligands", receptors_path = "./receptors", dockings_path = "./run_vina_run"):
    
    ligands_path = os.path.abspath(ligands_path)
    receptors_path = os.path.abspath(receptors_path)
    dockings_path = os.path.abspath(dockings_path)
    
    """These assume that:
        in receptor path the tree is:
            receptor1/
            receptor2/
            receptor3/
            (NO more directories)
        And inside of this directories there are three files:
            receptor1.pdb, step5_input.gro, step5_input.pdb and MEMBRANE.pdb
            
        As you see, this assumes that the receptor1.pdb has the same name as
        the parental directory (receptor1).
        
        in ligand path the tree is:
            ligand1/
            ligand2/
            ligand3/
            (NO more directories)
        And inside of this directories there are four files
        list.lis, is the list to change the order of the ligand, this is only the path, 
            In further step this file need to be created
            mol.frcmod
            mol.mol2
            mol.pdb
            list.lis
        
        in docking path the tree is the one generated with the script in palmiche/examples/Vina_Docking
        receptor1
            ligandX
                chainA
                    result1
                    .
                .   
            .    
        .
        """
    receptor_dict = {}
    for directory in tools.list_if_dir(receptors_path):
        
        receptor_dict[directory] = {"protein": os.path.join(receptors_path, directory, f"{directory}.pdb"),
                                    "step5_input_gro": os.path.join(receptors_path, directory, "step5_input.gro"),
                                    "step5_input_pdb": os.path.join(receptors_path, directory, "step5_input.pdb"),
                                    "membrane": os.path.join(receptors_path, directory, "MEMBRANE.pdb")}
    
   
    ligand_dict = {}
    for directory in tools.list_if_dir(ligands_path):
        ligand_dict[directory] = {"frcmod": os.path.join(ligands_path, directory, "mol.frcmod"),
                                  "mol2": os.path.join(ligands_path, directory, "mol.mol2"),
                                  "pdb": os.path.join(ligands_path, directory, "mol.pdb"),
                                  "list": os.path.join(ligands_path, directory, "list.lis")}
    
    docking_dict = {}
    for receptor in tools.list_if_dir(dockings_path):
        for ligand in tools.list_if_dir(os.path.join(dockings_path, receptor)): #In this way if the receptor where teasted to different ligands  (een so, they must be in the deffinition of the ligands_dict) this will tak it intop account
            for chain in tools.list_if_dir(os.path.join(dockings_path, receptor, ligand)):
                #Creating nested dictionary
                docking_dict.setdefault(receptor,{}).setdefault(ligand,{})[chain] = {"BE":os.path.join(dockings_path, receptor, ligand, chain, f"{ligand}_{chain}_BE.pdbqt"),
                                                                                     "NO":os.path.join(dockings_path, receptor, ligand, chain, f"{ligand}_{chain}_NO.pdbqt"),
                                                                                     "PO":os.path.join(dockings_path, receptor, ligand, chain, f"{ligand}_{chain}_PO.pdbqt")
                                                                                     }
    return receptor_dict, ligand_dict, docking_dict

def fix_pdb(step5_input_pdb, out = "PROTEIN.pdb"):
    """
    Take a pdb file and return the protein with the chains rectified, also convert CHARMM's atoms name to AMBER

    Parameters
    ----------
    step5_input_gro :  name of the file or path to the configuration file
        DESCRIPTION.A configuration file exported by CHARMM-GUI, gro or pdb.
        It is useful to use the last gro file because this one has the information
        for the construction of the gmx box on further steps.
    out : The name or path toe the output rectified pdb
 

    Returns
    -------
    1- PDB_OBJECT, this is a PDB object
    2- Also, a file called PROTEIN.pdb will be exported,
    This pdb file is protonated, with the chain rectified and also the element column added

    """
    if not os.path.isfile(step5_input_pdb):
        raise FileNotFoundError(f"The file {step5_input_pdb} doesn't exist or is not accessible")
    
    tmppdb1 = tempfile.NamedTemporaryFile(suffix='.pdb')
    tmppdb2 = tempfile.NamedTemporaryFile(suffix='.pdb')
    tmpndx = tempfile.NamedTemporaryFile(suffix='.ndx')

    ext = step5_input_pdb.split(".")[-1]
    if ext == "gro":
        tools.run(f"export GMX_MAXBACKUP=-1; echo gmx editconf -f {step5_input_pdb} -o {tmppdb1.name}")
        
        with open("warning.txt", "w") as w:
            w.write(f"""
            WARNING! 
            This is a warning from the function fix_pdb of assembleMD module
            The file {step5_input_pdb} is a gro file and
            gmx editconf was used to convert to pdb. If this function was used
            to start the assemble from CHARMM-GUI, please select the
            step5_input.pdb. Because the gro file doesn't have the information
            related with the chains and is not possible to fix because the residues
            number is changed in CHARMM-GUI to a continues list and doesn't use
            the real residues number of the pdb"
            """)
    else:
        tools.cp(step5_input_pdb, tmppdb1.name)
    
    #Keeping only the protein
    tools.run(f"""
            export GMX_MAXBACKUP=-1
            echo "q" | gmx make_ndx -f {tmppdb1.name} -o {tmpndx.name}
            echo "Protein" | gmx editconf -f {tmppdb1.name} -o {tmppdb2.name} -n {tmpndx.name}
          """)
    
    #************Rectify atom names, chains, residue...************************
    #This are the translations dict the kkey are the charmm nomenclature and the
    #values are the amber nomenclature for the atoms of the patching groups
    #ACE: acetylated N-terminus
    #CT3_NME: methylamidated C-terminus
    #The lst dict is for rename the HIS resnames, the former will change the atoms,
    #the residue and also the residue number
    #I also need to repair the chain
    ACE_ACE = {"CAY":"CH3", "HY1":"HH31", "HY2":"HH32", "HY3":"HH33", "CY":"C", "OY":"O"}
    CT3_NME = {"NT":"N", "HNT":"H", "CAT":"CH3", "HT1":"HH31", "HT2":"HH32", "HT3":"HH33"}
    HSD_HIS = {"HSD":"HIS"}
    
    
    PDB_OBJECT = pdb.PDB(tmppdb2.name)
    PDB_OBJECT.fix_chains()
    PDB_OBJECT.add_element()
    
    atoms = PDB_OBJECT.atoms
    #Get the Chains
    
    for atom in atoms:
        
    
        if atom.name in ACE_ACE:
            atom.name = ACE_ACE[atom.name]
            atom.resName = "ACE"
            #-1 Because this is for the N terminus that start in the PDB, this could
            #change, take a look to your system. This is not completely optimal. 
            #In the case that the aa start at 0you will get a -1 of residue number. 
            #However, why to cap if you have th first residue?
            #There is not need to change the name of the residue
            atom.resSeq -= 1
        elif atom.name in CT3_NME:
            atom.name = CT3_NME[atom.name]
            atom.resName = "NME"
            atom.resSeq += 1
        elif atom.resName in HSD_HIS:
            atom.resName = HSD_HIS[atom.resName]
        else:
            pass
        
    #This automatically convert to list
    PDB_OBJECT.write(tmppdb2.name, backup = False)
    #Get correct protonation, here I use amber99sb-ildn because it is by default in GROMACS
    
    #Probably a better idea is use CHARMM force fields in irder to dont worry about the atom names, but this use (amber) is consistent with my project 
    pka2gmx(tmppdb2.name,
            output = tmppdb2.name, 
            protein_forcefield = "amber99sb-ildn",
            water_forcefield = "tip3p",
            ph = 7.0,
            pKa = "propka3",
            pKa_file_out = False,
            itp_top_out = False,
            num_file = False)
    PDB_OBJECT = pdb.PDB(tmppdb2.name) #Need to load the new pdb

    PDB_OBJECT.write(out, backup = False)
    
    return PDB_OBJECT

def get_cryst1(step5_input_gro_path):
    tools.run(f"""
              gmx editconf -f {step5_input_gro_path} -o tmp.pdb
              """)
    TMP_PDB = pdb.PDB("tmp.pdb")
    #I need to dived by 10 because the pdb is in Angstrom
    box_vector = (TMP_PDB.cryst1.a/10, TMP_PDB.cryst1.b/10, TMP_PDB.cryst1.c/10)
    box_angles = (TMP_PDB.cryst1.alpha, TMP_PDB.cryst1.beta, TMP_PDB.cryst1.gamma)
    tools.rm("tmp.pdb")
    return box_vector, box_angles
    
    
def gmxmemb(step5_input_pdb_path,
            membrane_out_path,
            membrane_select = "(group System and ! group TIP3) and ! group Protein",
            lipid_ff_path = '/home/ale/MY_PYTHON_PACKEGES/palmiche/data/GROMACS.ff/Slipids_2020.ff'):
    """
    

    Parameters
    ----------
    step5_input_pdb_path : string path
        DESCRIPTION: The path to the final step of CHARMM-GUI, pdb file 
    membrane_out_path : string path
        DESCRIPTION: The path where you want to save the membrane.
        To get the membrane the following selection was used:
    membrane_select : string, gromacs selection
        DESCRIPTION: To get the membrane the following selection is used by default:
        (group System and ! group TIP3) and ! group Protein
        Therefore, this imply that your system only has:
            TIP3P, protein and the membrane, if you have ions and other molecules, you will need to change the selection properly.       
    lipid_ff_path : TYPE, optional: string path
        DESCRIPTION. The default is '/home/ale/MY_PYTHON_PACKEGES/palmiche/data/GROMACS.ff/Slipids_2020.ff'.
                Path to the lipid force field
    Returns
    -------
    None.

    """
    # Working with directories and files
    cwd = os.getcwd()
    tmpdir = tempfile.TemporaryDirectory()
    tools.cp(lipid_ff_path, tmpdir.name, r=True)
    lipid_ff_name = os.path.basename(lipid_ff_path).split('.ff')[0]
    step5_input_pdb_path = os.path.abspath(step5_input_pdb_path)
    membrane_out_path = os.path.abspath(membrane_out_path)
    
    os.chdir(tmpdir.name)

    #Create the selection
    with open("ndx.opt", "w") as opt:
        opt.write(f"\"MEMB\" {membrane_select};")
    tools.run(f"""
              echo "q" | gmx make_ndx -f {step5_input_pdb_path} -o tmp.ndx
              gmx select -s {step5_input_pdb_path} -sf ndx.opt -n tmp.ndx -on index
              """)
    
    #deleting the line _f0_t0.000 in the file
    with open("index.ndx", "r") as index:
        data = index.read()
        data = data.replace("_f0_t0.000","")
    with open("index.ndx", "w") as index:
        index.write(data)
    
    tools.run(f"""
                export GMX_MAXBACKUP=-1;         
                echo "MEMB"| gmx editconf -f {step5_input_pdb_path} -o tmpmemb.pdb -n index.ndx
                gmx pdb2gmx -f tmpmemb.pdb -o {membrane_out_path} -water none -ff {lipid_ff_name}
                """)
    os.chdir(cwd)

def protonate(mol_path, out_ext = "xyz", obabel = "openbabel.obabel"):
    """
    For this function you need openbabel
    could be installed as:
    sudo snap install openbabel
    pip install openbabel
    depending which one you have or if you installed using other method is how you call it.
    By default the call to openbable will be openbabel.obabel, but you could change this with the keyword obabel
    """
    
    if os.path.isfile(mol_path):    
        mol_name = os.path.basename(mol_path).split(".")[0]
        mol_ext = os.path.basename(mol_path).split(".")[-1]
        if mol_ext == "pdbqt":
            #No idea, but from pdbqt to xyz the H are not placed
            tmpfile = tempfile.NamedTemporaryFile(suffix='.pdb')
            tools.run(f"""
                      {obabel} {mol_path} -O {tmpfile.name} -h
                      {obabel} {tmpfile.name} -O {mol_name}.{out_ext} -h
                      """)
            tools.rm(tmpfile.name)
        else:
            tools.run(f"{obabel} {mol_path} -O {mol_name}.{out_ext} -h")
    else:
        raise FileNotFoundError(f"The file {mol_path} doesn't exist or is not accessible")

def gen_list_order(input_mol2_path, ref_mol2_path, list_mol2_out_path = "list.lis", pymol_path_cmd = "pymol", text_editor = "gedit"):
    """
    This function will open two python sessions and a text session.
    When the list had been created, then the python program will continue
    vim it is not so recommended because you will not get the help, but it will work;
    the command : "set number" will be usefull to see the number and create the list for vim.

    Parameters
    ----------
    
    input_mol2_path : string path
        DESCRIPTION. The molecule to change
    ref_mol2_path : string path
        DESCRIPTION.The reference molecule
    list_mol2_out_path : string path
        DESCRIPTION. The default is "list.lis".
        A text file with the correct oreder.
        If in the reference molecule the atoms are in the oreder A, B, C. And in the
        input the atoms are in the order B, C, A. then the list must
        have these three lines: 3, 1, 2. What means: the position
        of the number (first line, second line, and so on) represent 
        the atom order list of the reference molecule. The numbers
        on the line, represent where the reference atoms are in the 
        input structure. The file must have just the lines with the
        specification numbers.
    pymol_path_cmd : TYPE, optional
        DESCRIPTION. The default is "pymol".

    Returns
    -------
    None.

    """
    if not os.path.isfile(input_mol2_path):
        raise FileNotFoundError(f"The file {input_mol2_path} doesn't exist or is not accessible")
    if not os.path.isfile(ref_mol2_path):
        raise FileNotFoundError(f"The file {ref_mol2_path} doesn't exist or is not accessible")
    
    string = f"""
    As you see, there are two pymol sessions and a gedit sesion:
        1- The pymol session with the mol {os.path.basename(input_mol2_path)}
        is the molecule that we want to change the atom order.
        2- The pymol session with the mol {os.path.basename(ref_mol2_path)}
        is the reference molecule.
    You need to set the "label > atom identifier > ID" in pymol
    
    If in the reference molecule the atoms are in the oreder A, B, C. And in the
    input the atoms are in the order B, C, A. then the list must
    have these three lines: 3, 1, 2. What means: the position
    of the number (first line, second line, and so on) represent 
    the atom order list of the reference molecule. The numbers
    on the line, represent where the reference atoms are in the 
    input structure. The file must have just the lines with the
    specification numbers.
    
    See the following example:
    Reference molecule                      Input molecule
          5                                     9
          |                                     |
      10--1--6                              11--10--8
          |                                     |
      11--2--7                              14--5--12
          |                                     |
      12--3--8                               7--1--6
          |                                     |
      13--4--9                              4--13--3
          |                                     |
          14                                    2
    Then the list.lis will have the following lines. In parenthesis the line number:
        (1)10
        (2)5
        (3)1
        (4)13
        (5)9
        (6)8
        (7)12
        (8)6
        (9)3
        (10)11
        (11)14
        (12)7
        (13)4
        (14)2
    
    
    
    When the {text_editor} session is closed the python programme will continue, not before.
              """
    print(string)
    tools.run(f"""
              {pymol_path_cmd} {input_mol2_path}  &
              {pymol_path_cmd} {ref_mol2_path} &
              {text_editor} {list_mol2_out_path}
              """)
    
    

def order_mol2(input_mol2_path, ref_mol2_path, list_mol2_path, lname_mol2 = 'UNL', obabel = "openbabel.obabel"):

    """
    

    Parameters
    ----------
    input_mol2_path : string, path
        The molecule that you want to change the atom list
        order in any openbabel supported format. Is important that 
        it has the same number of atoms that the reference. Check
        for the H atoms, if the implicit H are missing, run first
        obabel mol.abc -O mol.xyz -h
        and compare the structures in Pymol in order to 
        create the list file.
        
    ref_mol2_path : string, path
        The molecule that you will use as reference for the 
        atom list order in mol2 format
    
    list_mol2_path : string, path
        A text file with the correct oreder. If in the ref
        structure the atoms are in the oreder A, B, C. And in the
        input the atoms are in the order B, C, A. then the list must
        have these three lines: 3, 1, 2. What means: the position
        of the number (first line, second line, and so on) represent 
        the atom order list of the reference structure. The numbers
        on the line, represents where the reference atom is in the 
        input structure. The file must have just the lines with the
        specification   numbers.
    
    lname_mol2 : string, optional
       The three letter identifier for the ligand. Default is UNL.

    obabel : string, optional
       The name for execute openbabel in case that the extension of input_mol2_path is different from .xyz . Default is openbabel.obabel.
    Returns
    -------
    None.

    """
    file_input_mol2 = os.path.basename(input_mol2_path)
    out_path = os.path.dirname(input_mol2_path)
    
    #print(os.path.exists(input_mol2_path), input_mol2_path)
    if file_input_mol2.split('.')[-1] == 'xyz':
        to_work = input_mol2_path
    else:
        xyztmp = tempfile.NamedTemporaryFile(suffix=".xyz")
        tools.run(f'{obabel} {input_mol2_path} -O {xyztmp.name}')
        to_work = xyztmp.name
    
    order = [int(i) - 1 for i in open(list_mol2_path, 'r').readlines()] # -1 is because the python array start at 0
    
    with open(to_work, 'r') as f:
        lines_input = f.readlines()
    Natoms = int(lines_input[0])
    coords = lines_input[2:2+Natoms]

    if len(coords) != len(order): exit(f"the len of coords in {input_mol2_path} doesn't match with the number of line in {ref_mol2_path}")
    
    update_coords = []
    for i in order:
        update_coords.append(coords[i])
    update_coords = [c.split()[1:] for c in update_coords]
    
    with open(ref_mol2_path, 'r') as f:
        lines_ref = f.readlines()
    name_final_out = file_input_mol2.split('.')[0]+".mol2"
    
    #The file is written to the directory where the input is
    with open(os.path.join(out_path, name_final_out), 'wt') as final_out:
        i = 0
        while i < len(lines_ref):
            final_out.write(lines_ref[i])
            if lines_ref[i].startswith('@<TRIPOS>MOLECULE'):
                final_out.write(f'    {lname_mol2}\n')
                i += 1
            if lines_ref[i].startswith('@<TRIPOS>ATOM'):
                for j in range(len(update_coords)):
                    
                    atom_split = lines_ref[i+j+1].split()
                    to_print = "\t".join(atom_split[:2]) + "\t" + "\t".join(update_coords[j]) + '\t' + "\t".join(atom_split[-4:-2]) + "\t" + lname_mol2 + "\t" + atom_split[-1] + '\n'   
                    
                    final_out.write(to_print)
                i = i + Natoms
            i += 1  

def run_tleap(mol2_path, frcmod_path):
    name = os.path.basename(mol2_path).split(".")[0]    
    tmpin = tempfile.NamedTemporaryFile(suffix='.in')

    with open(tmpin.name, 'w') as t:
        t.write(f"""
                  loadAmberParams {frcmod_path}
                  A = loadMol2 {mol2_path}
                  saveAmberParm A {name}.prmtop {name}.crd
                  quit
                  """)
        
    tools.run(f"""
               tleap -s -f {tmpin.name} > {name}_tleap.out
               acpype -p {name}.prmtop -x {name}.crd
               """)
    
    [tools.run(f"obabel {gro} -O {gro.split('.')[0]}.pdb") for gro in glob("*_GMX.gro")]



def toptleap2itp(topology, out_prefix = "UNL"):
    sections = tools.get_top_sections(topology)
    
    #Getting the important sections, in a dict is much more easy, but we
    #have the problem of the duplication of dihedrals section
    for section in sections:
        if section[0] == "atoms":
            atom_section = section
        elif section[0] == "atomtypes":
            atomtypes_section = section
        else:
            pass
    
    #I need to run again the loop because I am not able to pop or delete on 
    #the fly, because the for loop will raise error, I am not able to sort
    #Because I dont know the implications of it.
    itp_sections = ""
    for section in sections:
        if section[0] not in ["defaults" , "atomtypes", "system", "molecules"]:
            itp_sections += f"[ {section[0]} ]\n{''.join(section[1])}"
        

    #Writing the atomtypes itp
    with open("atomtypes.itp", "w") as at:
        at.write(f"[ {atomtypes_section[0]} ]\n{''.join(atomtypes_section[1])}")
        
    #Writing the general itp
    #Even when the five ligand on different chains will have the same
    #contents on this file. You need to make a distinction because the
    #moleculetype definition inside of the topology file that references
    #the name of the molecule. And this is needed for the pulling simulations
    
    include_posre = f"""
; Include Position restraint file for heavy atoms
#ifdef POSRES
#include "{out_prefix}_posre.itp"
#endif
"""
    with open(f"{out_prefix}.itp", "w") as itp:
        itp.write(f"{itp_sections}{include_posre}")
    
    #creating the posre, first we will look for the heavy atoms
    heavy_atoms = []
    for atom in atom_section[1]:
        if not atom.startswith(";"):#This is necessary because the commented lines in gromacs topology files
            split = atom.strip().split()#strip for the case of blank lines
            if split:#The list has something, this is all for the case of blank lines
                if split[4][0].lower() == "h": #This is the column that has the atom name, I select the first char because could be H0, HH13, etc, and all of them are Hydrogens atoms, hoping that all of them start with H
                    heavy_atoms.append(split[0])#Here is the column that has the atom index
            
    #Writing the posre file
    with open(f"{out_prefix}_posre.itp", "w") as posre:
        posre.write("[ position_restraints ]\n")
        posre.write(f"{';i':<5}{'funct':^5}{'fcx':^20}{'fcy':^20}{'fcz':^20}\n")
        for atom_index in heavy_atoms:
            posre.write(f"{atom_index:^5}{1:^5}{'POSRES_LIG':^20}{'POSRES_LIG':^20}{'POSRES_LIG':^20}\n")
                
        
def topoltop(topol, numb_lipids, out_name = None, backup = True):
    """
    !!!!!This function is NOT general at all, it is just created here for stetic
    and is strong depended of what I needed to do in my project, so, use it
    with precaution

    Parameters
    ----------
    topol : TYPE, path, string
        DESCRIPTION. a path or name to a GROMACS topology file

    out_name : TYPE, string
        DESCRIPTION. The name of the output, if None, teh topol will be used
        It is just the name, not the path, the path will be the path were the topol file is

    Returns
    -------
    None.
    the modified topology

    """               
    ff_include_old = """
; Include forcefield parameters
#include "./amber99sb-star-ildn.ff/forcefield.itp"
"""
    ff_include_new = """
; Include forcefield parameters
#include "./amber99sb-star-ildn.ff/forcefield.itp"

; Include forcefield parameters		
#include "./Slipids_2020.ff/forcefield.itp"

; Include GAFF2 atom definition
#include "./GAFF2.ff/atomtypes.itp"
"""

    rest_str_old = """
; Include water topology
#include "./amber99sb-star-ildn.ff/tip3p.itp"

#ifdef POSRES_WATER
; Position restraint for each water oxygen
[ position_restraints ]
;  i funct       fcx        fcy        fcz
   1    1       1000       1000       1000
#endif

; Include topology for ions
#include "./amber99sb-star-ildn.ff/ions.itp"

[ system ]
; Name
Protein

[ molecules ]
; Compound        #mols
Protein_chain_A     1
Protein_chain_B     1
Protein_chain_C     1
Protein_chain_D     1
Protein_chain_E     1"""

    rest_str_new = f"""
; Include POPC chain topology
#include "toppar/POPC.itp"

; position restraints for POPC lipid
#ifdef POSRES
[ position_restraints ]
   20     1     0.0             0.0            POSRES_FC_LIPID
#endif

#ifdef DIHRES
[ dihedral_restraints ]
   25    36    28    30     1   -120.0      2.5       DIHRES_FC
   60    63    65    67     1      0.0      0.0       DIHRES_FC
#endif	

; Include ligand topologies
#include "toppar/LIA.itp"
#include "toppar/LIB.itp"
#include "toppar/LIC.itp"
#include "toppar/LID.itp"
#include "toppar/LIE.itp"

; Include water topology
#include "./amber99sb-star-ildn.ff/tip3p.itp"

#ifdef POSRES_WATER
; Position restraint for each water oxygen
[ position_restraints ]
;  i funct       fcx        fcy        fcz
   1    1       1000       1000       1000
#endif

; Include topology for ions
#include "./amber99sb-star-ildn.ff/ions.itp"

[ system ]
; Name
Protein

[ molecules ]
; Compound        #mols
Protein_chain_A     1
Protein_chain_B     1
Protein_chain_C     1
Protein_chain_D     1
Protein_chain_E     1
POPC		{numb_lipids}
LIA		1
LIB		1
LIC		1
LID		1
LIE		1"""
   
    with open(topol, "r") as t:
        topol_data = t.read()
    topol_data = topol_data.replace(ff_include_old, ff_include_new)
    topol_data = topol_data.replace(rest_str_old, rest_str_new)
    topol_data = topol_data.replace("topol_Protein_chain_", "toppar/topol_Protein_chain_")
       
    if out_name:
        with open(os.path.join(os.path.dirname(topol), out_name), "w") as out:
            out.write(topol_data)
    else:
        if backup: tools.backoff(topol)
        with open(topol, "w") as out:
            out.write(topol_data)

def posre_BB_SC(pdb_path):
    tools.run(f"""
    subs='1e+08'
    
    echo "Backbone" | gmx genrestr -f {pdb_path} -o backbone_posre -fc $subs $subs $subs
    sed "s/$subs/POSRES_FC_BB/g" backbone_posre.itp > tmp ; mv tmp backbone_posre.itp
    
    echo "SideChain-H" | gmx genrestr -f {pdb_path} -o sidechain_posre -fc $subs $subs $subs
    sed "s/$subs/POSRES_FC_SC/g" sidechain_posre.itp > tmp ; mv tmp sidechain_posre.itp
              """)
    
def include_posre_itp(itp_in, *args, itp_out = None, backup =True):
    """
    

    Parameters
    ----------
    itp_in : string path
        DESCRIPTION: .itp file form GROMACS
    *args : string
        DESCRIPTION: You must give at least one,
        These are the file that you will included with the position restrain.
        See the example below:
           
            ; Include Position restraint file for protein backbone
            #ifdef POSRES
            #include "{arg}"
            #endif
            
    itp_out : TYPE, optional, string path
        DESCRIPTION. The default is None. The path to the output file, if None,
        the original file will be updated and the former backed it up

    Raises
    ------
    KeyError
        DESCRIPTION. In case that you don't give extra arguments

    Returns
    -------
    None.

    """
    if not len(args):
        raise KeyError("You need to provied at least on argument to fill in the itp")
    
    include_str = ""
    for arg in args:
        include_str += f"""
; Include Position restraint file for protein
#ifdef POSRES
#include "{arg}"
#endif
        """
     
    with open(itp_in, "r") as f:
        itp_lines = f.readlines()
    for (i,line) in enumerate(reversed(itp_lines)):
        if "; Include Position restraint file" in line:
            break
    #Here in this line are bassically eliminating the position restraing definition
    #And replaced by a new one
    itp_lines = "".join(itp_lines[:len(itp_lines) - 1 - i]) + include_str
    
    if itp_out:
        with open(itp_out, "w") as f:
            f.write(itp_lines)
    else:
        if backup: tools.backoff(itp_in)
        with open(itp_in, "w") as f:
            f.write(itp_lines)       
def gmxsolvate(abs_path, input_file, topology, editconf_box, editconf_angles = (90,90,60), editconf_bt = "tric", solvate_cs = "spc216", out_file = "solvated.pdb"):
    """
    

    Parameters
    ----------
    abs_path : TYPE, string path
        DESCRIPTION: Here we must deffine the path were all the necessary files
        are together, topologies, forcefields, configurations files, etc
    input_file : TYPE, string
        DESCRIPTION. Name of the configuration input file
    topology : TYPE, string
        DESCRIPTION. GROMACS topology file
    editconf_box : TYPE tuple
        DESCRIPTION. HEre we deffine the components of the vector on editconf (a,b,c)
    editconf_angles : TYPE, optional, tuple
        DESCRIPTION. The default is (90,90,60). This default value is for
        hexagonal boxes. idela for membrane simulations
    editconf_bt : TYPE, optional, string
        DESCRIPTION. The default is "tric". Type of GROMACS box
    solvate_cs : TYPE, optional, string
        DESCRIPTION. The default is "spc216". Configuration for the solvent
    out_file : TYPE, optional, string
        DESCRIPTION. The default is "solvated.pdb". Name of the solvated box

    Returns
    -------
    None.

    """
    
    initial_cwd = os.getcwd()
    
    os.chdir(abs_path)
    tools.run(f"""
                export GMX_MAXBACKUP=-1
                gmx editconf -f {input_file} -o {out_file} -bt {editconf_bt} -box {' '.join([str(i) for i in editconf_box])} -angles {' '.join([str(i) for i in editconf_angles])}
                gmx solvate -cp {out_file} -p {topology} -cs {solvate_cs} -o {out_file}
              """)
             
    os.chdir(initial_cwd)

def gmxions(abs_path, input_file, topology, out_file = "ions.pdb", genion_pname = "K", genion_nname = "CL", genion_rmin = 1.0):
    """
    

    Parameters
    ----------
    abs_path : TYPE
        DESCRIPTION.
    input_file : TYPE, string 
        DESCRIPTION. the name of the input file inside the abs_path
    topology : TYPE, string
        DESCRIPTION. the name of the topology file inside the abs_path
    out_file : TYPE, optional, string
        DESCRIPTION. The default is "ions.pdb". The name of the output file
        inside the abs_path. Path will not handlead.
        
        For the rest of the parameters see the fromacs documentation of genion
    genion_pname : TYPE, optional
        DESCRIPTION. The default is "K".
    genion_nname : TYPE, optional
        DESCRIPTION. The default is "CL".
    genion_rmin : TYPE, optional
        DESCRIPTION. The default is 1.0.

    Returns
    -------
    None.

    """
    initial_cwd = os.getcwd()
    tmptpr = tempfile.NamedTemporaryFile(suffix = '.tpr')
    tmpmdp = tempfile.NamedTemporaryFile(suffix = '.mdp')
    os.chdir(abs_path)
    tools.run(f"""
                export GMX_MAXBACKUP=-1
                gmx grompp -f {tmpmdp.name} -c {input_file} -p {topology} -o {tmptpr.name}
                echo "SOL" | gmx genion -s {tmptpr.name} -p {topology} -o {out_file} -neutral -pname {genion_pname} -nname {genion_nname} -rmin {genion_rmin}
              """)
    # Cleanning
    tools.rm("mdout.mdp")
    os.chdir(initial_cwd)

def finalndx(abs_path, input_file, ndxout = "index.ndx", chainsID = ["A", "B", "C", "D", "E"]):
    
    initial_cwd = os.getcwd()
    os.chdir(abs_path)
    
    tmpopt = tempfile.NamedTemporaryFile(suffix='.opt')
    tmpndx = tempfile.NamedTemporaryFile(suffix='.ndx')
    #Nice use of gmx select, see the use of the parenthesis
    with open(tmpopt.name, "w") as opt:
        opt.write(f"""
                  "MEMB" ((group System and ! group Water_and_ions) and ! group Protein) and ! ({"resname LI"+" or resname LI".join(chainsID)});
                  "SOLU" group Protein or {"resname LI"+" or resname LI".join(chainsID)};
                  "SOLV" group Water_and_ions;
                  """)
    tools.run(f"""
                export GMX_MAXBACKUP=-1
                echo "q" | gmx make_ndx -f {input_file} -o {tmpndx.name}
                gmx select -s {input_file} -sf {tmpopt.name} -n {tmpndx.name} -on {ndxout}
                """)
    
    #deleting the line _f0_t0.000 in the file
    with open(ndxout, "r") as index:
        data = index.read()
        data = data.replace("_f0_t0.000","")
    with open(ndxout, "w") as index:
        index.write(data)
    
    
    os.chdir(initial_cwd)

def GROMACS_SMAUG_JOB(job_name, conf_file_name = "ASSEMBLE.pdb", 
                               partition="deflt", 
                               output="myjob.out", error="myjob.err",
                               cpus_per_task=12, gpus = 0, time="2-00:00",
                               nodes=1, nice=0, version = "2021.1", exclude=None):
    template = f"""#!/bin/bash
#SBATCH --partition {partition}
#SBATCH --output {output}
#SBATCH --error {error}
#SBATCH --cpus-per-task={cpus_per_task}
#SBATCH --time {time}
#SBATCH --job-name={job_name}
#SBATCH --nodes={nodes}
#SBATCH --nice={nice}
#SBATCH --gpus={gpus}
    """
    if exclude:
        template += f"#SBATCH --exclude={exclude}\n"
    template += f"""
\n\n
# This block is echoing some SLURM variables
echo "Job execution start: $(date)"
echo "JobID = $SLURM_JOBID"
echo "Host = $SLURM_JOB_NODELIST"
echo "Jobname = $SLURM_JOB_NAME"
echo "Subcwd = $SLURM_SUBMIT_DIR"
echo "SLURM_TASKS_PER_NODE = $SLURM_TASKS_PER_NODE"
echo "SLURM_CPUS_PER_TASK = $SLURM_CPUS_PER_TASK"
echo "SLURM_CPUS_ON_NODE = $SLURM_CPUS_ON_NODE"

cd $(pwd)
source /data/shared/opt/gromacs/{version}/bin/GMXRC

export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:

# Minimization

gmx grompp -f step6.0_minimization.mdp -o step6.0_minimization.tpr -c {conf_file_name} -r {conf_file_name} -p topol.top -n index.ndx
gmx mdrun -v -deffnm step6.0_minimization

# Equilibration
cnt=1
cntmax=6

for ((i=${{cnt}}; i<${{cntmax}}+1; i++)); do
	if [ $i == "1" ]; then
		gmx grompp -f step6.${{i}}_equilibration.mdp -o step6.${{i}}_equilibration.tpr -c step6.$[${{i}}-1]_minimization.gro -r {conf_file_name} -p topol.top -n index.ndx
		gmx mdrun -v -deffnm step6.${{i}}_equilibration -nt 12
	else
		gmx grompp -f step6.${{i}}_equilibration.mdp -o step6.${{i}}_equilibration.tpr -c step6.$[${{i}}-1]_equilibration.gro -r {conf_file_name} -p topol.top -n index.ndx
		gmx mdrun -v -deffnm step6.${{i}}_equilibration -nt 12
	fi
done 
    """
    return(template)

if __name__ == "__main__":
    pass 

            
        
